from __future__ import annotations

from typing import Union, Optional
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, LetterCase, Undefined
from collections import UserList

import wxcadm
from wxcadm import log
from .common import *


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Number:
    phone_number: Optional[str] = None
    """ The phone number """
    extension: Optional[str] = None
    """ The extension """
    routing_prefix: Optional[str] = None
    """ The Location routing prefix """
    esn: Optional[str] = None
    """ The Enterprise Number """
    state: Optional[str] = None
    """ The state of the number """
    phone_number_type: Optional[str] = None
    """ The type of phone number """
    included_telephony_types: Optional[str] = None
    """ The type of telephony that is included with this number """
    main_number: Optional[bool] = None
    """ Whether this is a Main Number for a Location """
    toll_free_number: Optional[bool] = None
    """ Whether the number is a toll-free number """
    _location: Optional[Union[dict, wxcadm.Location]] = None
    _owner: Optional[Union[dict, wxcadm.Person, wxcadm.VirtualLine, wxcadm.Workspace]] = None
    org: Optional[wxcadm.Org] = field(init=False, repr=False)

    @property
    def location(self) -> Optional[Union[wxcadm.Location, dict]]:
        if self._location:
            found_location = self.org.locations.get(id=self._location['id'])
            if found_location is not None:
                return found_location
            else:
                return self._location
        else:
            return None
        return None

    @property
    def owner(self):
        if self._owner:
            owner_id = self._owner.get('id', None)
            if owner_id is not None:
                owner_type = self._owner['type']
                if owner_type == 'PEOPLE':
                    owner = self.org.people.get(id=owner_id)
                    if owner is not None:
                        return owner
                    else:
                        return self._owner
                elif owner_type == 'PLACE':
                    owner = self.org.workspaces.get(id=owner_id)
                    if owner is not None:
                        return owner
                    else:
                        return self._owner
                elif owner_type == 'VIRTUAL_LINE':
                    owner = self.org.virtual_lines.get(id=owner_id)
                    if owner is not None:
                        return owner
                    else:
                        return self._owner
                elif owner_type == 'CALL_QUEUE':
                    owner = self.org.call_queues.get(id=owner_id)
                    if owner is not None:
                        return owner
                    else:
                        return self._owner
                elif owner_type == 'HUNT_GROUP':
                    owner = self.org.hunt_groups.get(id=owner_id)
                    if owner is not None:
                        return owner
                    else:
                        return self._owner
                elif owner_type == 'AUTO_ATTENDANT':
                    owner = self.org.auto_attendants.get(id=owner_id)
                    if owner is not None:
                        return owner
                    else:
                        return self._owner
                elif owner_type == 'PAGING_GROUP':
                    for group in self.org.paging_groups:
                        if group.id == owner_id:
                            return group
                    return self._owner
                elif owner_type == 'VOICEMAIL_GROUP':
                    for group in self.org.voicemail_groups:
                        if group.id == owner_id:
                            return group
                    return self._owner
                else:
                    log.warning(f'Unknown owner type "{owner_type}"')
                    return self._owner
            else:
                if self._owner['type'] == 'VOICE_MESSAGING':
                    return self.location.voice_portal
        else:
            return None

    def activate(self):
        """ Activate the phone number

        Returns:
            Number: The updated :class:`Number` with the new state

        """
        webex_api_call('put', f'v1/telephony/config/locations/{self.location.id}/numbers',
                       params={'orgId': self.org.id},
                       payload={'phoneNumbers': [self.phone_number]})
        return True


class NumberList(UserList):
    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location]):
        super().__init__()
        self.parent = parent
        if isinstance(parent, wxcadm.Org):
            self.org: wxcadm.Org = parent
            self.data: list = self._get_data()
        elif isinstance(parent, wxcadm.Location):
            self.org: wxcadm.Org = parent.parent
            self.data: list = self._get_data(parent)
        else:
            raise ValueError('parent much be an Org or Location')

    def _get_data(self, location: Optional[wxcadm.Location] = None) -> list:
        data = []
        params = {'orgId': self.org.id}
        if location is not None:
            params['locationId'] = location.id
        response = webex_api_call('get', 'v1/telephony/config/numbers', params=params)
        for number in response['phoneNumbers']:
            this_number: Number = Number.from_dict(number)
            this_number.org = self.org
            data.append(this_number)
        return data

    def refresh(self):
        """ Refresh the list from Webex """
        if isinstance(self.parent, wxcadm.Location):
            self.data = self._get_data(self.parent)
        else:
            self.data = self._get_data()
        return self

    def get(self,
            phone_number: Optional[str] = None,
            extension: Optional[str] = None,
            esn: Optional[str] = None,
            state: Optional[str] = None,
            location: Optional[wxcadm.Location] = None
            ) -> Optional[Union[Number, list]]:
        """ Find a Number or list of Number entries matching the given criteria.

        If ``'phone_number'`` or ``'esn'`` is specified, then the response will be a :class:`Number` instance or None
        if no match is found. If ``'extension'`` and ``'location'`` are specified, the response will be a
        :class:`Number` if that extension matches within the provided location. If ``'extension'`` is specified
        without a ``'location'``, a :class:`Number` instance will be returned if there is only one match. A list will
        be returned if there are multiple matches. All others will return a list or None if there are no matches.

        Args:
            phone_number (str, optional): The phone number to match.
            extension (str, optional): The extension to match.
            esn (str, optional): The ESN to match.
            state (str, optional): The state to match.
            location (wxcadm.Location, optional): The Location to match.

        Returns:
            wxcadm.Number: The matching number
            list[wxcadm.Number]: The matched numbers

        """
        if phone_number is None and extension is None and esn is None and state is None and location is None:
            raise ValueError('A parameter is required')
        # Handle the single-value searches first
        for number in self.data:
            if phone_number is not None and phone_number in number.phone_number:
                return number
            if esn is not None and esn == number.esn:
                return number
            # If an extension and a location were provided, the user is expecting a single entry
            if extension is not None and location is not None:
                if extension == number.extension and location == number.location:
                    return number
        # Then do the list-return searches
        if state is not None or location is not None or extension is not None:
            result = []
            for number in self.data:
                # Extensions can exist in more than one location
                if extension is not None and extension == number.extension:
                    result.append(number)
                if state is not None and state == number.state:
                    result.append(number)
                if location is not None and location.id == number.location.id:
                    result.append(number)
            # To handle legacy extension searches, only return the Number if it's the only one in the result
            if len(result) == 1:
                return result[0]
            else:
                return result
        return None

    def get_by_owner(self, owner):
        for number in self.data:
            if number.owner == owner:
                return number

    def add(self,
            location: wxcadm.Location,
            numbers: list,
            number_type: Optional[str] = 'DID',
            state: Optional[str] = 'ACTIVE'):
        """ Add a list of numbers to a Location

        .. note::

        This is only supported for Locations using Local Gateway or Non-integrated PSTN as the PSTN type.

        Args:
            location (Locaton): The :class:`Location` to add the numbers to
            numbers (list): A list of numbers, as strings
            number_type (str, optional): The type of numbers being added. `'TOLLFREE'` and `'DID'` are supported.
                Defaults to `DID`
            state (str, optional): The state to put the numbers in. Valid values are `'ACTIVE'` and `'INACTIVE'`.
                Defaults to `'ACTIVE'` 
            
        Returns:
            NumberList: The updated :class:`NumberList` with the newly added numbers

        Raises:
            wxcadm.APIError: Raised when the number add is rejected by Webex

        """
        payload = {
            'phoneNumbers': numbers,
            'numberType': number_type,
            'state': state
        }
        webex_api_call('post',
                       f'v1/telephony/config/locations/{location.id}/numbers',
                       params={'orgId': self.org.id},
                       payload=payload)
        self.refresh()
        return self

    def validate(self, numbers: list):
        """ Validate a list of phone numbers prior to adding.

        Args:
            numbers (list): A list of phone numbers, each as a string

        Returns:
            dict: The validation response from Webex

        """
        response = webex_api_call('post', f'v1/telephony/config/actions/validateNumbers/invoke',
                                  params={'orgId': self.org.id}, payload={'phoneNumbers': numbers})
        return response



from __future__ import annotations

from dataclasses import dataclass, field
from collections import UserList
import wxcadm
from wxcadm import log

import requests
from typing import Optional
from .exceptions import *


class RedSky:
    def __init__(self, username: str, password: str):
        """ Initialize a connection to RedSky and obtain basic account info

        Args:
            username (str): The Horizon admin username
            password (str): The Horizon admin password

        """
        # Log into Horizon and get access token info
        payload = {"username": username, "password": password}
        r = requests.post("https://api.wxc.e911cloud.com/auth-service/login", json=payload)
        if r.ok:
            response = r.json()
            log.debug(response)
        else:
            raise APIError("Cannot log into RedSky Horizon")
        self._access_token = response['accessToken']
        self.org_id = response['userProfileTO']['company']['id']
        self._refresh_token = response['refreshTokenInfo']['id']
        self._full_config = response
        self._buildings = []
        self._users = None

    @property
    def _headers(self):
        headers = {"Authorization": "Bearer " + self._access_token}
        return headers

    def _token_refresh(self):
        """A method to refresh the access_toke using the stored refresh_token"""
        r = requests.post("https://api.wxc.e911cloud.com/auth-service/token/refresh",
                          headers=self._headers,
                          json=self._refresh_token)
        if r.status_code == 200:
            response = r.json()
            self._access_token = response['accessToken']
            self._refresh_token = response['refreshTokenInfo']['id']
        else:
            raise APIError("There was an error refreshing the RedSky token")

    def get_all_locations(self):
        """ Get all the Locations that RedSky knows about

        .. note ::

            This method is only useful in very specific reporting cases, since you only get the Location information
            and no device-level information.

        Returns:
             list[dict]: A list of the location information as a dict

        """
        response: dict[str, list[dict]] = {'corporate': [], 'personal': []}
        more_data = True
        page = 1
        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/geography-service/locations/parent/{self.org_id}",
                             headers=self._headers, params={'page': page})
            if r.status_code == 401:
                self._token_refresh()
                continue
            response['corporate'].extend(r.json())
            for user in self.users:
                response['personal'].extend(user.user_locations)
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1
        return response

    @property
    def buildings(self):
        """A list of all the RedSkyBuilding instances within this RedSky account

        Returns:
            list[RedSkyBuilding]: The RedSkyBuilding instances

        """
        self._buildings = []
        more_data = True
        page = 1
        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/geography-service/buildings/parent/{self.org_id}",
                             params={'page': page, 'pageSize': 100, 'searchTerm': None, 'origin': 'default'},
                             headers=self._headers)
            if r.status_code == 401:
                if r.status_code == 401:
                    self._token_refresh()
                    continue
            if r.status_code == 200:
                response = r.json()
                for item in response:
                    building = RedSkyBuilding(self, item)
                    self._buildings.append(building)
            else:
                raise APIError("Something went wrong getting the list of buildings")
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1
        return self._buildings

    def get_building_by_name(self, name: str):
        """Get the RedSkyBuilding instance for a given name

        Args:
            name (str): The name of the Building to return. Not case-sensitive.

        Returns:
            RedSkyBuilding: The instance of the RedSkyBuilding class. None is returned if no match is found.

        """
        for building in self.buildings:
            if building.name.lower() == name.lower():
                return building

        return None

    def get_building_by_webex_location(self, location: wxcadm.Location):
        """Get the RedSkyBuilding instance for a given Location instance.

        This method will first try to match on the Location ID in Webex, which will work if the Building was created
        by ***wxcadm**. If no match is found, it will attempt to match on the Location Name. The Webex Location name and
        the RedSky Building name must match exactly, although case is ignored.

        Args:
            location (Location): The Location instance to search against.

        Returns:
            RedSkyBuilding: The instance of the RedSkyBuilding class. None is returned if no match is found.

        """
        buildings = self.buildings
        for building in buildings:
            if building.supplemental_data == location.id[-20:]:
                return building
        for building in buildings:
            if building.name.lower() == location.name.lower():
                return building
        return None

    def add_building(self, webex_location: Optional[wxcadm.Location] = None,
                     address_string: Optional[str] = None, create_location: bool = True):
        """ Add a new Building to RedSky

        The ``create_location`` arg defaults to True, which will automatically create a RedSkyLocation called "Default".
        If you want to prevent a Location from being created, pass ``create_location=False``.

        Either a ``webex_location`` or ``address_string`` argument should be provided. If both are provided, the Webex
        Location instance will take precedence

        Args:
            webex_location (Location): The Webex Location instance to create the building for
            address_string (str, optional): The complete address, as a string
            create_location (bool. optional): Whether to create a RedSkyLocation called "Default" in the RedSkyBuilding

        Returns:
            RedSkyBuilding: The RedSkyBulding instance that was created

        Raises:
            ValueError: Raised when trying to create a RedSkyBuilding for a Location that is not in the U.S.
            wxcadm.exceptions.APIError: Raised on all API failures

        """
        if webex_location is not None:
            if webex_location.address['country'] != "US":
                raise ValueError("Cannot create Building for non-US location")

            a1 = webex_location.address['address1']
            a2 = webex_location.address.get("address2", "")
            city = webex_location.address['city']
            state = webex_location.address['state']
            zip = webex_location.address['postalCode']
            full_address = f"{a1} {a2}, {city}, {state} {zip}"
        else:
            full_address = address_string

        payload = {"name": webex_location.name,
                   "supplementalData": webex_location.id[-20:],
                   "parentOrganizationUnitId": self.org_id,
                   "fullAddress": full_address,
                   "origin": "default"
                   }
        log.debug(payload)
        r = requests.post("https://api.wxc.e911cloud.com/geography-service/buildings",
                          headers=self._headers, json=payload)
        if r.ok:
            building = self.get_building_by_webex_location(webex_location)
            if create_location is True:
                building = self.get_building_by_webex_location(webex_location)
                if building is not None:
                    building.add_location("Default", ecbn=webex_location.main_number.replace("+1", ""))
                else:
                    raise APIError("The created Building cannot be located. Location not created.")
            return building
        else:
            raise APIError(f"Something went wrong adding Building: {r.text}")

    @property
    def held_devices(self):
        """All the HELD devices known to RedSky"""
        more_data = True
        page = 1
        held_devices = []
        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/admin-service/held/org/{self.org_id}",
                             headers=self._headers,
                             params={'page': page, 'pageSize': 100, 'searchTerm': None, 'type': None})
            if r.status_code == 401:
                self._token_refresh()
                continue
            if r.status_code == 200:
                response = r.json()
                for device in response:
                    held_devices.append(device)
            else:
                raise APIError(f"Something went wrong getting the HELD devices {r.text}")
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1
        return held_devices

    def phones_without_location(self):
        """Get a list of phone (HELD) devices that don't have a RedSkyLocation associated

        Returns:
            list[dist]: A list of the device properties from RedSky

        """
        devices = []
        for device in self.held_devices:
            if device['erl'] is None and device['deviceType'] == "HELD":
                devices.append(device)
        return devices

    def clients_without_location(self):
        """Get a list of soft client (HELD+) devices that don't have a RedSkyLocation associated

        Returns:
            list[dict]: A list of the device properties from RedSky

        """
        devices = []
        for device in self.held_devices:
            if device['erl'] is None and device['deviceType'] == "HELD_PLUS":
                devices.append(device)
        return devices

    def get_mac_discovery(self, mac: Optional[str] = None) -> list | dict | None:
        """ Get the current MAC address mapping defined in RedSky Horizon

        When called without the ``mac`` argument, all current MAC mappings will be returned. When a ``mac`` is passed,
        the entry for that MAC will be returned. If no match can be found, None will be returned.

        Args:
            mac (str, optional): The MAC value to get the details for

        Returns:
            list[dict]: A list of all the MAC address mappings. If a ``mac`` argument was passed, the dict of that
            entry is returned, or None is returned if no match was found

        Raises:
            wxcadm.exceptions.APIError: Raised on any error from the RedSKy API

        """
        mappings = []
        more_data = True
        page = 1
        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/networking-service/macAddress/company/{self.org_id}",
                             headers=self._headers, params={'page': page, 'pageSize': 100})
            if r.ok:
                response = r.json()
                for item in response:
                    mappings.append(item)
            else:
                if r.status_code == 401:
                    self._token_refresh()
                    continue
                else:
                    raise APIError(f"There was a problem getting MAC mapping: {r.text}")
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1

        if mac is not None:
            log.info(f"Finding MAC Discover for MAC: {mac.upper()}")
            for entry in mappings:
                if entry['macAddress'].upper() == mac.upper():
                    log.debug(f"Match found: {entry['id']}")
                    return entry
            log.warning("No MAC match found")
            return None
        return mappings

    def add_mac_discovery(self, mac: str, location: "RedSkyLocation", description: str = ""):
        """ Add a new MAC address mapping to RedSky Horizon

        Args:
            mac (str): The MAC address to add. RedSky isn't picky about formatting or case, so any MAC format should
                work
            location (RedSkyLocation): The RedSkyLocation instance to add the mapping to
            description (str, optional): A description of the device or any other information to store

        Returns:
            dict: The configuration of the mapping after processing by RedSky

        Raises:
            wxcadm.exceptions.APIError: Raised on any error from the RedSKy API

        """
        payload = {"macAddress": mac,
                   "locationId": location.id,
                   "orgId": self.org_id,
                   "description": description}
        r = requests.post(f"https://api.wxc.e911cloud.com/networking-service/macAddress",
                          headers=self._headers, json=payload)
        if r.ok:
            added_mapping = r.json()
        else:
            raise APIError(f"There was a problem adding the mapping: {r.text}")
        return added_mapping

    def delete_mac_discovery(self, mac: Optional[str] = None, entry_id: Optional[str] = None) -> bool:
        """ Delete a MAC Address mapping from Horizon Mobility

        If you have found the MAC Discovery entry yourself using the :py:meth:`get_mac_discovery()` method, you
        can pass the ``['id']`` of that entry as the ``entry_id``. If you pass the ``mac`` argument, that lookup will
        be done for you.

        Either the ``mac`` or the ``entry_id`` argument must be passed. If both are passed, the ``entry_id`` will
        take precedence and the ``mac`` value will be ignored.

        Args:
            mac (str, optional): The MAC address to delete
            entry_id (str, optional): The ``['id']`` of the MAC Discovery entry dic

        Returns:
            bool: True on success, False otherwise

        Raises:
            ValueError: Raised when no ``mac`` or ``entry_id`` is present in the arguments

        """
        if mac is None and entry_id is None:
            raise ValueError("Either mac or entry_id argument must be present")
        log.info(f"Deleting MAC Discovery: {mac}")

        if entry_id is None:
            entry = self.get_mac_discovery(mac)
            if entry is None:
                return False
            else:
                entry_id = entry['id']

        r = requests.delete(f"https://api.wxc.e911cloud.com/networking-service/macAddress/{entry_id}",
                            headers=self._headers,
                            params={"companyId": self.org_id})
        log.info(f"Response Code: {r.status_code}")
        if r.ok:
            return True
        else:
            return False

    def get_lldp_discovery(self):
        """ Get the current LLDP chassis and port mappings defined in RedSky Horizon

        Returns:
            list[dict]: A list of all the LLDP mappings

        Raises:
            APIError: Raised on any error from the RedSKy API

        """
        mappings = []
        more_data = True
        page = 1
        page_limit = 100
        while more_data is True:
            log.debug("Getting more data")
            r = requests.get(f"https://api.wxc.e911cloud.com/networking-service/networkSwitch/company/{self.org_id}",
                             headers=self._headers, params={'page': page, "pageSize": page_limit})
            log.debug(f"GET Response Headers: {r.headers}")
            if r.ok:
                response = r.json()
                for item in response:
                    item['ports'] = []
                    port_more_data = True
                    port_page = 1
                    port_page_limit = 100
                    while port_more_data is True:
                        rport = requests.get(f"https://api.wxc.e911cloud.com/networking-service/networkSwitchPort/"
                                             f"networkSwitch/{item['id']}",
                                             headers=self._headers,
                                             params={'page': port_page, 'pageSize': port_page_limit})
                        if rport.ok:
                            ports = rport.json()
                            for port in ports:
                                item['ports'].append(port)
                        else:
                            raise APIError(f"There was a problem getting Chassis Ports: {rport.text}")
                        port_more_data = port_page < int(rport.headers.get('X-Pagination-Count', 1))
                        port_page += 1
                    mappings.append(item)
            else:
                if r.status_code == 401:
                    self._token_refresh()
                    continue
                else:
                    raise APIError(f"There was a problem getting Chassis mapping: {r.text}")
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1
        return mappings

    def get_lldp_discovery_by_chassis(self, chassis_id: str):
        """ Get the LLDP chassis and port mappings for a give Chassis identifier

        Args:
            chassis_id (str): The Chassis identifier

        Returns:
            dict: The LLDP discovery configuration. None is returned if no match is found

        """
        for entry in self.get_lldp_discovery():
            if entry['chassisId'].upper() == chassis_id.upper():
                return entry
        return None

    def add_lldp_discovery(self, chassis: str, location: "RedSkyLocation", ports: list = None, description: str = ""):
        """ Add a new LLDP mapping to RedSky Horizon

        If only a chassis identifier is provided, the chassis will be added. If a list of ports are provided, the method
        will determine if the chassis already exists. If it does, the ports will be added and the chassis mapping will
        not be changed. If the chassis does not exist, it will be added and mapped to the provided RedSkyLocation, and
        the list of ports will be added, mapped to the same RedSkyLocation.

        Args:
            chassis (str): The chassis identifier
            location (RedSkyLocation): The RedSkyLocation instance to add the mapping to
            ports (list, optional): A list of ports (i.e. ['1','2','3'] or ['B000B4BB1BF4:P1'] to add mapping for
            description (str, optional): A description of the device or any other information to store

        Returns:
            dict: The configuration of the mapping after processing by RedSky

        Raises:
            wxcadm.exceptions.APIError: Raised on any error from the RedSKy API

        """
        if ports is not None:
            # Determine if the chassis already exists
            chassis_exists = False
            chassis_id = ""
            current_lldp = self.get_lldp_discovery()
            for entry in current_lldp:
                if entry['chassisId'].lower() == chassis.lower():
                    chassis_id = entry['id']
                    chassis_exists = True
            if chassis_exists is False:
                new_chassis = self._add_chassis_discovery(chassis, location, description)
                chassis_id = new_chassis['id']
            for port in ports:
                self._add_port_discovery(chassis_id, port, location, description)

            # Now that we added everything, let's get the details of the new chassis and its ports
            current_lldp = self.get_lldp_discovery()
            for entry in current_lldp:
                if entry['chassisId'].lower == chassis.lower():
                    return entry
        else:
            new_chassis = self._add_chassis_discovery(chassis, location, description)
            return new_chassis

    def delete_lldp_chassis(self, chassis_id: str, delete_ports: bool = False):
        """ Delete an LLDP mapping from Horizon Mobility

        Args:
            chassis_id (str): The chassis ID of the LLDp Discovery to delete
            delete_ports (bool, optional): Whether to delete all ports first. Defaults to False

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Deleting Chassis Discovery: {chassis_id}")
        chassis_entry = self.get_lldp_discovery_by_chassis(chassis_id)

        if delete_ports is True:
            log.info("Deleting all ports")
            for port in chassis_entry['ports']:
                log.info(f"\tDeleting Port: {port['portId']}")
                r = requests.delete(f"https://api.wxc.e911cloud.com/networking-service/networkSwitchPort/{port['id']}",
                                    headers=self._headers,
                                    params={
                                        "companyId": self.org_id
                                    })
                log.debug(f"Response Code: {r.status_code}")

        r = requests.delete(f"https://api.wxc.e911cloud.com/networking-service/networkSwitch/{chassis_entry['id']}",
                            headers=self._headers,
                            params={
                                "companyId": self.org_id
                            })
        log.debug(f"Response Code: {r.status_code}")

        if r.ok:
            return True
        else:
            return False

    def update_lldp_location(self, entry_id: str, chassis: str, new_location: RedSkyLocation, description: str):
        payload = {
            'chassisId': chassis,
            'companyId': self.org_id,
            'id': entry_id,
            'locationId': new_location.id,
            'description': description
        }
        r = requests.put("https://api.wxc.e911cloud.com/networking-service/networkSwitch",
                         headers=self._headers, json=payload)
        if r.ok:
            return True
        else:
            raise APIError(f"There was a problem updating the chassis: {r.text}")

    def update_lldp_port_location(self, entry_id: str, chassis_id: str, port: str,
                                  new_location: RedSkyLocation, description: str):
        payload = {
            'description': description,
            'erlId': new_location.id,
            'id': entry_id,
            'networkSwitchId': chassis_id,
            'portId': port
        }
        r = requests.put("https://api.wxc.e911cloud.com/networking-service/networkSwitchPort",
                         headers=self._headers, json=payload)
        if r.ok:
            return True
        else:
            raise APIError(f"There was a problem updating the chassis: {r.text}")

    def _add_chassis_discovery(self, chassis, location, description=""):
        payload = {"chassisId": chassis,
                   "companyId": self.org_id,
                   "locationId": location.id,
                   "description": description}
        r = requests.post("https://api.wxc.e911cloud.com/networking-service/networkSwitch",
                          headers=self._headers, json=payload)
        if r.ok:
            response = r.json()
        else:
            raise APIError(f"There was a problem adding the chassis: {r.text}")
        return response

    def _add_port_discovery(self, chassis_id, port, location, description=""):
        payload = {"erlId": location.id,
                   "portId": port,
                   "description": description,
                   "networkSwitchId": chassis_id}
        r = requests.post(f"https://api.wxc.e911cloud.com/networking-service/networkSwitchPort",
                          headers=self._headers, json=payload)
        if r.ok:
            response = r.json()
        else:
            raise APIError(f"There was a problem adding the port to the chassis: {r.text}")
        return response

    def get_bssid_discovery(self, bssid: Optional[str] = None) -> list | dict | None:
        """ Get the current BSSID mappings defined in RedSky Horizon

        When called without the ``bssid`` argument, all current BSSID mappings will be returned. When a ``bssid`` is
        passed, the entry for that BSSID will be returned. If no match can be found, None will be returned.

        Args:
            bssid (str, optional): The BSSID value to get the details for

        Returns:
            list[dict]: A list of all the BSSID mappings

        Raises:
            wxcadm.exceptions.APIError: Raised on any error from the RedSKy API

        """
        log.info("Getting BSSID Discovery")
        mappings = []
        more_data = True
        page = 1
        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/networking-service/bssid/company/{self.org_id}",
                             headers=self._headers, params={'page': page, 'pageSize': 100})
            if r.ok:
                response = r.json()
                for item in response:
                    mappings.append(item)
            else:
                if r.status_code == 401:
                    self._token_refresh()
                    continue
                else:
                    raise APIError(f"There was a problem getting BSSID mapping: {r.text}")
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1

        if bssid is not None:
            log.info(f"Finding BSSID Discovery for BSSID: {bssid.upper()}")
            for entry in mappings:
                if entry['bssid'].upper() == bssid.upper():
                    log.debug(f"Match found: {entry['id']}")
                    return entry
            log.warning("No BSSID match found")
            return None
        return mappings

    def add_bssid_discovery(self, bssid: str, location: "RedSkyLocation", description: str = "", masking: bool = True):
        """ Add a new BSSID mapping to RedSky Horizon

        Args:
            bssid (str): The BSSID to add. RedSky isn't picky about formatting or case, so any MAC format should work
            location (RedSkyLocation): The RedSkyLocation instance to add the mapping to
            description (str, optional): A description of the device or any other information to store
            masking (bool, optional): Whether to apply the default last-digit masking to the BSSID

        Returns:
            dict: The configuration of the mapping after processing by RedSky

        Raises:
            wxcadm.exceptions.APIError: Raised on any error from the RedSKy API

        """
        payload = {"bssid": bssid,
                   "bssidMasking": masking,
                   "locationId": location.id,
                   "orgId": self.org_id,
                   "description": description}
        r = requests.post(f"https://api.wxc.e911cloud.com/networking-service/bssid",
                          headers=self._headers, json=payload)
        if r.ok:
            added_mapping = r.json()
        else:
            raise APIError(f"There was a problem adding the mapping: {r.text}")
        return added_mapping

    def delete_bssid_discovery(self, bssid: Optional[str] = None, entry_id: Optional[str] = None) -> bool:
        """ Delete a BSSID Mapping from Horizon Mobility

        If you have found the BSSID Discovery entry yourself using the :py:meth:`get_bsssid_discovery()` method, you
        can pass the ``['id']`` of that entry as the ``entry_id``. If you pass the ``bssid`` argument, that lookup will
        be done for you.

        Either the ``bssid`` or the ``entry_id`` argument must be passed. If both are passed, the ``entry_id`` will
        take precedence and the ``bssid`` value will be ignored.

        Args:
            bssid (str, optional): The BSSID to delete
            entry_id (str, optional): The ``['id']`` of the BSSID Discovery entry dict

        Returns:
            bool: True on success, False otherwise

        Raises:
            ValueError: Raised when no ``bssid`` or ``entry_id`` is present in the arguments

        """
        if bssid is None and entry_id is None:
            raise ValueError("Either bssid or entry_id argument must be present")
        log.info(f"Deleting BSSID Discovery: {bssid}")

        if entry_id is None:
            entry = self.get_bssid_discovery(bssid)
            if entry is None:
                return False
            else:
                entry_id = entry['id']
        r = requests.delete(f"https://api.wxc.e911cloud.com/networking-service/bssid/{entry_id}",
                            headers=self._headers,
                            params={"companyId": self.org_id})
        log.info(f"Response Code: {r.status_code}")
        if r.ok:
            return True
        else:
            return False

    def get_ip_range_discovery(self, ip_start: Optional[str] = None,
                               ip_end: Optional[str] = None,
                               range_for_ip: Optional[str] = None,
                               type: Optional[str] = 'private') -> list | dict | None:
        """ Get the current IP Range mappings defined in Horizon Mobility

        This method can be used to retrieve Private IP Range Discovery (the default) as well as Public IP Ranges.
        Public IP Ranges can be retrieved by adding ``type='public'`` to the arguments.

        When this method is called without an argument, all IP Range Discovery entries will be returned as a list, or an
        empty list if there are none defines. When passed with any argument, the method will return the dict of the
        matching entry, or None will be returned if there is no match.

        The ``range_for_ip`` argument takes priority and will return the entry of the range that will include the
        given IP. For example, with an existing entry from 192.168.1.1 - 192.168.1.255, a
        ``range_for_ip='192.168.1.100'`` will return that entry, to make finding ranges by included IPs easier. If this
        argument is passed, the ``ip_start`` and ``ip_end`` arguments will be ignored.

        If both ``ip_start`` and ``ip_end`` are given, an entry that matches either value will be returned. If only one
        is passed, the entry must mach that value.

        Args:
            ip_start (str, optional): The first IP in the range
            ip_end (str, optional): The last IP in the range
            range_for_ip (str, optional): An IP to find the matching range for (i.e. an IP within an IP Range entry)
            type (str, optional): Accepted values are ``'private'`` and ``'public'``. Defaults to ``'private'``.


        Returns:
            list[dict]: A list of all the IP Range mappings. An empty list is returned when no argument is passed and
            there are no entries in the IP Range Discovery. If a filter argument was passed, the entry will be the dict
            of the matching entry or None if no match was found.

        Raises:
            ValueError: Raised if ``type`` argument is not ``'private'`` or ``'public'``
            wxcadm.exceptions.APIError: Raised on any error from the RedSKy API

        """
        if type.lower() != 'private' and type.lower() != 'public':
            raise ValueError("Unrecognized type value")

        if type.lower() == 'private':
            endpoint = 'ipRange'
        elif type.lower() == 'public':
            endpoint = 'publicIpRange'
        mappings = []
        more_data = True
        page = 1
        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/networking-service/{endpoint}/company/{self.org_id}",
                             headers=self._headers, params={'page': 1, 'pageSize': 100})
            if r.ok:
                response = r.json()
                for item in response:
                    mappings.append(item)
            else:
                if r.status_code == 401:
                    self._token_refresh()
                    continue
                else:
                    raise APIError(f"There was a problem getting IP Range mapping: {r.text}")
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1

        if range_for_ip is not None:
            for entry in mappings:
                if entry['ipAddressLow'] <= range_for_ip <= entry['ipAddressHigh']:
                    return entry
            return None

        if ip_start is not None or ip_end is not None:
            for entry in mappings:
                if entry['ipAddressLow'] == ip_start or entry['ipAddressHigh'] == ip_end:
                    return entry
            return None
        return mappings

    def add_ip_range_discovery(self, ip_start: str, ip_end: str, location: "RedSkyLocation", description: str = ""):
        """ Add a new IP Range mapping to RedSky Horizon

        Args:
            ip_start (str): The first IP in the range to add
            ip_end (str): The last IP in the range to add
            location (RedSkyLocation): The RedSkyLocation instance to add the mapping to
            description (str, optional): A description of the device or any other information to store

        Returns:
            dict: The configuration of the mapping after processing by RedSky

        Raises:
            wxcadm.exceptions.APIError: Raised on any error from the RedSKy API

        """
        payload = {"ipAddressLow": ip_start,
                   "ipAddressHigh": ip_end,
                   "locationId": location.id,
                   "description": description}
        r = requests.post(f"https://api.wxc.e911cloud.com/networking-service/ipRange",
                          headers=self._headers, json=payload, params={'companyId': self.org_id})
        if r.ok:
            added_mapping = r.json()
        else:
            raise APIError(f"There was a problem adding the mapping: {r.text}")
        return added_mapping

    def add_public_ip_range(self, ip_start: str, ip_end: str, description: str = ""):
        """ Add a new Public IP Range to RedSky Horizon

        .. note::

            Public IP Ranges are not mapped to a Location. They are to ensure that only requests from recognized
            corporates networks are accepted by Horizon Mobility

        Args:
            ip_start (str): The first IP in the range
            ip_end (str): The last IP in the range
            description (str, optional): A description for the Public IP Range entry

        Returns:
            dict: The created entry

        """
        payload = {"ipAddressLow": ip_start,
                   "ipAddressHigh": ip_end,
                   "description": description,
                   "companyId": self.org_id}
        r = requests.post(f"https://api.wxc.e911cloud.com/networking-service/publicIpRange",
                          headers=self._headers, json=payload, params={'companyId': self.org_id})
        if r.ok:
            added_range = r.json()
        else:
            raise APIError(f"There was a problem adding the mapping: {r.text}")
        return added_range

    def delete_ip_range_discovery(self, ip_start: str, ip_end: str):
        """ Delete an IP Range mapping from Horizon Mobility

        The ``ip_start`` and ``ip_end`` arguments must match the IP Range Discovery entry to delete

        Args:
            ip_start (str): The first IP in the range to delete
            ip_end (str): The last IP in the range to delete

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Deleting IP Range Discovery: {ip_start} - {ip_end}")
        entry = self.get_ip_range_discovery(ip_start=ip_start, ip_end=ip_end)
        if entry is None:
            return False
        r = requests.delete(f"https://api.wxc.e911cloud.com/networking-service/ipRange/{entry['id']}",
                            headers=self._headers,
                            params={"companyId": self.org_id})
        log.info(f"Response Code: {r.status_code}")
        if r.ok:
            return True
        else:
            return False

    @property
    def users(self):
        self._users = RedSkyUsers(parent=self)
        return self._users


class RedSkyUsers(UserList):
    def __init__(self, parent: RedSky):
        log.info("Initializing RedSkyUsers instance")
        super().__init__()
        self.parent = parent
        self.data = []

        # Handle RedSky pagination
        more_data = True
        next_page = 1
        page_size = 100

        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/admin-service/deviceUser/company/{self.parent.org_id}",
                             headers=self.parent._headers, params={'page': next_page, 'pageSize': page_size})
            if r.ok:
                data = r.json()
                log.info("Getting RedSky Users")
                log.debug(f"\tPage: {r.headers['X-Pagination-Page']} of {r.headers['X-Pagination-Count']}")
                for user in data:
                    this_user = RedSkyUser(parent=self.parent,
                                           id=user['id'],
                                           email=user['heldUserId'],
                                           raw=user)
                    self.data.append(this_user)
                # Figure out if we need to get more data or not
                if int(r.headers['X-Pagination-Page']) < int(r.headers['X-Pagination-Count']):
                    more_data = True
                    next_page += 1
                else:
                    more_data = False
                    log.debug(f"\tGot {len(self.data)} Users")
            else:
                raise APIError(f"There was a problem getting the RedSky Users")

    def get_by_email(self, email: str) -> Optional[RedSkyUser]:
        for user in self.data:
            if user.email == email:
                return user
        return None


@dataclass
class RedSkyUser:
    parent: RedSky
    id: str
    email: str
    raw: field(repr=False, default_factory=dict)

    @property
    def user_locations(self):
        """ Get the user-defined locations

        These are the locations that the user has entered manually in the Webex app (or the MyE911 app, if they have
        used it)

        Returns:
            list[dict]: List of location information
        """
        log.info("Getting RedSky User locations")
        log.debug(f"User ID: {self.id}")
        r = requests.get(f"https://api.wxc.e911cloud.com/geography-service/locations/deviceUser/{self.id}",
                         headers=self.parent._headers, params={'page': 1, 'pageSize': 100})
        if r.ok:
            data = r.json()
            return data
        else:
            log.error(f"There was a problem getting the RedSky User locations: {r.text}")
            raise APIError(f"There was a problem getting the User's locations")

    @property
    def devices(self):
        """ Get the list of all Webex app devices for the user

        Because the user email is only associated to soft clients, not desk phones, the "devices" returned are only
        computers that have communicated an Emergency Response Location to RedSky. Unfortunately, with the Webex app,
        due to privacy restrictions, it can be difficult to know which device the user is current logged in to.

        Returns:
            list[dict]: A list of the device dictionaries, including the ERL for the device

        """
        log.info(f"Getting devices for RedSky User {self.email}")
        r = requests.get(f"https://api.wxc.e911cloud.com/admin-service/device",
                         headers=self.parent._headers,
                         params={'deviceUserId': self.id})
        if r.ok:
            return r.json()
        else:
            return False


class RedSkyBuilding:
    """A RedSky Horizon Building"""

    def __init__(self, parent: RedSky, config: dict):
        """Initialize a RedSkyBuilding instance.

        Args:
            parent (RedSky): The RedSky instance that this Building belongs to
            config: (dict): The dict returned by the RedSky API containing the Building information

        """
        self._parent: RedSky = parent
        self._raw_config: dict = config
        self.id: str = config.get("id")
        """The ID of the Building"""
        self.name: str = config.get("name")
        """The name of the Building"""
        self.supplemental_data: str = config.get("supplementalData", None)
        """Supplemental data for the Building"""
        self.type: str = config.get("type", "unknown")
        """The type of Building"""
        self.address: dict = config.get("address")
        """The physical address of the building"""
        self._locations: Optional[list] = None

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.id

    @property
    def locations(self):
        """A list of RedSkyLocation instances associated with this Building

        Returns:
            list[RedSkyLocation]: List of RedSkyLocation instances

        """
        self._locations = []
        more_data = True
        page = 1
        while more_data is True:
            r = requests.get(f"https://api.wxc.e911cloud.com/geography-service/locations/parent/{self.id}",
                             headers=self._parent._headers, params={'page':  page, 'pageSize': 100})
            if r.status_code == 200:
                response = r.json()
                for item in response:
                    location = RedSkyLocation(self, item)
                    self._locations.append(location)
            else:
                raise APIError(f"There was a problem getting the Locations for Building {self.name}")
            more_data = page < int(r.headers.get('X-Pagination-Count', 1))
            page += 1
        return self._locations

    def get_location_by_name(self, name: str):
        """ Return the RedSkyLocation instance that matches the name provided.

        Args:
            name (str): The name of the location to search for. Not case-sensitive

        Returns:
            RedSkyLocation: The instance of the RedSkyLocation class. None is returned if no match can be found

        """
        for location in self.locations:
            if location.name.lower() == name.lower():
                return location
        return None

    def add_location(self, location_name: str, ecbn: str = "", location_info: str = ""):
        payload = {"name": location_name,
                   "elin": ecbn,
                   "info": location_info,
                   "organizationUnitId": self.id}
        r = requests.post("https://api.wxc.e911cloud.com/geography-service/locations",
                          headers=self._parent._headers, json=payload)
        if r.ok:
            return True
        else:
            raise APIError("Something went wrong adding the Location: {r.text}")

    @property
    def bssid_discovery(self):
        """ The BSSID Discovery entries associated with Locations at this RedSkyBuilding """
        entries = []
        for location in self.locations:
            entries.extend(location.bssid_discovery)
        return entries

    @property
    def lldp_discovery(self):
        """ The LLDP Discovery entries associated with Locations at this RedSkyBuilding"""
        entries = []
        for location in self.locations:
            entries.extend(location.lldp_discovery)
        return entries

    @property
    def mac_discovery(self):
        """ The MAC Address Discovery entries associated with Locations at this RedSkyBuilding """
        entries = []
        for location in self.locations:
            entries.extend(location.mac_discovery)
        return entries

    @property
    def ip_range_discovery(self):
        """ The IP Range Discovery entries associated with Locations at this RedSkyBuilding """
        entries = []
        for location in self.locations:
            entries.extend(location.ip_range_discovery)
        return entries


class RedSkyLocation:
    """A RedSky Horizon Location"""

    def __init__(self, parent: RedSkyBuilding, config: dict):
        """Initialize a RedSkyLocation instance

        Args:
            parent (RedSkyBuilding): The RedSkyBuilding instance that this Location belongs to
            config (dict): The dict returned by the RedSky API containing the Location information

        """
        self._parent = parent
        self._raw_config = config
        self.id = config.get("id")
        self.name = config.get("name")
        self.address = config.get("address")
        self.type = config.get("type", "unknown")
        self.supplemental_data = config.get("supplementalData", None)
        self.org_name_override = config.get("orgNameOverride")
        self.info = config.get("info")
        self.address_entity_name: str = config.get("addressEntityName")
        self.elin: dict = config.get("elin", None)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.id

    @property
    def bssid_discovery(self):
        """ The BSSID Discovery entries associated with this RedSkyLocation """
        return [entry for entry in self._parent._parent.get_bssid_discovery() if entry['location']['id'] == self.id]

    @property
    def lldp_discovery(self):
        """ The LLDP Discovery entries associated with this RedSkyLocation"""
        return [entry for entry in self._parent._parent.get_lldp_discovery() if entry['location']['id'] == self.id]

    @property
    def mac_discovery(self):
        """ The MAC Address Discovery entries associated with this RedSkyLocation """
        return [entry for entry in self._parent._parent.get_mac_discovery() if entry['location']['id'] == self.id]

    @property
    def ip_range_discovery(self):
        """ The IP Range Discovery entries associated with this RedSkyLocation """
        return [entry for entry in self._parent._parent.get_ip_range_discovery() if entry['location']['id'] == self.id]



import requests.exceptions
from requests import Response
import xmltodict
from typing import Optional


class XSIResponse:
    """ An XSIResponse is returned by Webex for all XSI commands """
    def __init__(self, response: Response):
        try:
            self.raw_response = response.json()
        except requests.exceptions.JSONDecodeError:
            self.raw_response = response.text
        """ The raw response dict or text """
        self.summary: Optional[str] = None
        """ Summary text for errors """
        self.success: bool = False
        """ Whether or not the command was successful """

        if 'ErrorInfo' in self.raw_response:
            self.summary = self.raw_response['ErrorInfo']['summary']['$']
        if response.ok:
            self.success = True
        else:
            self.success = False

    def __bool__(self):
        """ True when self.success is True """
        if self.success is True:
            return True
        else:
            return False



from __future__ import annotations

from collections import UserList
from typing import Optional, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .person import Person
    from .workspace import Workspace

import wxcadm
from wxcadm import log
from .common import *


class DECTHandset:
    def __init__(self, dect_network: DECTNetwork, config: Optional[dict] = None):
        self.dect_network = dect_network
        self.id: Optional[str] = config.get('id', None)
        self.index: str = config.get('index', '')
        self.display_name: str = config.get('displayName', '')
        self.access_code: str = config.get('accessCode')
        self.lines: list = config.get('lines', [])
        self._mac: str = config.get('mac', None)

    @property
    def mac(self) -> str:
        if self._mac is None:
            response = webex_api_call("get",
                                      f"v1/telephony/config/locations/{self.dect_network.location_id}/"
                                      f"dectNetworks/{self.dect_network.id}/handsets/{self.id}")
            self._mac = response['mac'].upper()
        return self._mac

    def delete(self) -> bool:
        """ Delete the Handset

        Returns:
            bool: True on success

        """
        webex_api_call('delete',
                       f'v1/telephony/config/locations/{self.dect_network.location_id}/dectNetworks/'
                       f'{self.dect_network.id}/handsets/{self.id}')
        return True

    def set_handset_display_name(self, display_name: str) -> bool:
        """ Set/change the name displayed on the handset

        Args:
            display_name (str): The new display name

        Returns:
            bool: True on success, False otherwise

        """
        payload = {
            'line1MemberId': self.lines[0]['memberId'],
            'customDisplayName': display_name
        }
        if self.lines[1]:
            payload['line2MemberId'] = self.lines[1]['memberId']
        try:
            webex_api_call(
                'put',
                f'v1/telephony/config/locations/{self.dect_network.location_id}/dectNetworks'
                f'/{self.dect_network.id}/handsets/{self.id}',
                payload=payload)
        except wxcadm.APIError:
            return False
        self.display_name = display_name
        return True


class DECTBaseStation:
    def __init__(self, dect_network: DECTNetwork, config: Optional[dict] = None):
        self.dect_network: DECTNetwork = dect_network
        self.id: Optional[str] = config.get('id', None)
        self.mac: str = config.get('mac', '').upper()
        self.number_of_lines_registered: int = config.get('numberOfLinesRegistered', 0)
        log.info(f"Initializing DECTBaseStation with MAC {self.mac}")

    def delete(self) -> bool:
        """ Delete the Base Station

        Returns:
            bool: True on success

        Raises:
            wxcadm.APIError: Raised when the delete is rejected by Webex
        """
        webex_api_call(
            'delete',
            f'v1/telephony/config/locations/{self.dect_network.location_id}/dectNetworks/'
            f'{self.dect_network.id}/baseStations/{self.id}'
        )
        return True

    def get_handsets(self):
        """ List of :class:`DECTHandset` instances associated with this base station """
        handsets = []
        response = webex_api_call(
            'get',
            f'v1/telephony/config/locations/{self.dect_network.location_id}/dectNetworks/'
            f'{self.dect_network.id}/baseStations/{self.id}'
        )
        for handset in response['handsets']:
            this_handset = DECTHandset(dect_network=self.dect_network, config=handset)
            handsets.append(this_handset)
        return handsets


class DECTNetwork:
    def __init__(self,
                 location: Optional[wxcadm.Location] = None,
                 id: Optional[str] = None,
                 config: Optional[dict] = None):
        # If we were given an ID, even if there was a config present, go fetch the details, but we need the location
        if id is not None:
            if location is None:
                raise ValueError('location must be provided with id')
            else:
                log.debug(f'Getting DECT Network {id} at Location {location.id}')
                config = webex_api_call(
                    'get',
                    f'v1/telephony/config/locations/{location.id}/dectNetworks/{id}'
                )
        log.info(f"Initializing DECTNetwork with Name {config['name']}")
        self.id: str = config.get('id', None)
        self.name: str = config.get('name', '')
        self.display_name: str = config.get('displayName', '')
        self.chain_id: str = config.get('chainId', '')
        self.model: str = config.get('model', '')
        self.default_access_code_enabled: bool = config.get('defaultAccessCodeEnabled', False)
        self.default_access_code: str = config.get('defaultAccessCode', '')
        self.number_of_base_stations: int = config.get('numberOfBaseStations', 0)
        self.number_of_handsets_assigned: int = config.get('numberOfHandsetsAssigned', 0)
        self.number_of_lines: int = config.get('numberOfLines', 0)
        self.location_id: str = config['location'].get('id', None)

    @property
    def handsets(self) -> list:
        handsets = []
        response = webex_api_call(
            'get',
            f'v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}/handsets'
        )
        for handset in response['handsets']:
            this_handset = DECTHandset(self, config=handset)
            handsets.append(this_handset)
        return handsets

    def delete_base_station(self, base_station: Union[str, DECTBaseStation]) -> list:
        """ Delete a Base Station from this DECT Network.

        The `base_station` parameter can be either a :class:`DECTBaseStation` instance or the Base Station MAC address
        as a string.

        Args:
            base_station (str, DECTBaseStation: The :class:`DECTBaseStation` instance or the MAC address of the Base
            Station to delete

        Returns:
            list: The remaining list of Base Stations

        Raises:
            wxcadm.APIError: Raised when the delete is rejected by Webex

        """
        log.debug('Started DECTNetwork.delete_base_station()')
        if isinstance(base_station, DECTBaseStation):
            log.info(f'Deleting DECT Base Station: {base_station.id} ({base_station.mac.upper()}')
            base_station.delete()
        else:
            for base in self.base_stations:
                if base.mac.upper() == base_station.upper():
                    log.info(f'Deleting DECT Base Station: {base.id} ({base.mac.upper()}')
                    base.delete()
        return self.base_stations

    def delete(self) -> bool:
        """ Delete the DECT Network

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f'Deleting DECT Network {self.id}')
        try:
            webex_api_call(
                'delete',
                f'v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}'
            )
        except wxcadm.APIError:
            log.error(f'Webex API exception')
            return False
        else:
            log.debug('Success')
            return True

    def set_name(self, name: str, display_name: Optional[str] = None) -> bool:
        """ Set/change the name of the DECT Network

        Args:
            name (str): The new name of the DECT Network
            display_name (str, optional): The Display Name of the DECT Network. Only required if it needs to be changed

        Returns:
            bool: True on success, False otherwise

        """
        if display_name is not None:
            new_display_name = display_name
        else:
            new_display_name = self.display_name
        payload = {
            'name': name,
            'displayName': new_display_name,
            'defaultAccessCodeEnabled': self.default_access_code_enabled,
            'defaultAccessCode': self.default_access_code
        }
        try:
            webex_api_call(
                'put',
                f'v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}',
                payload=payload
            )
        except wxcadm.APIError:
            return False
        self.name = name
        return True

    def set_display_name(self, display_name: str) -> bool:
        """ Set/change the display name of the DECT Network

        Args:
            display_name (str): The new display name of the DECT Network

        Returns:
            bool: True on success, False otherwise

        """
        payload = {
            'name': self.name,
            'displayName': display_name,
            'defaultAccessCodeEnabled': self.default_access_code_enabled,
            'defaultAccessCode': self.default_access_code
        }
        try:
            webex_api_call(
                'put',
                f'v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}',
                payload=payload
            )
        except wxcadm.APIError:
            return False
        self.display_name = display_name
        return True

    def enable_default_access_code(self, access_code: Optional[str] = None):
        """ Enable the Default Access Code for handsets.

        If an Access Code has never been defined for the DECT Network, it should be provided using the
        `access_code` parameter. If an Access Code had previously been defined and disabled, it does not need to
        be provided, but can if desired. If a previous Access Code is not present and one is not provided, a
        ValueError will be raised.

        Args:
            access_code (str, optional): The Access Code to set (digits only)

        Returns:
            bool: True on success, False otherwise

        Raises:
            ValueError: Raised when `access_code` is not provided and a previous value does not exist

        """
        if access_code is None:
            if self.default_access_code is None or self.default_access_code == '':
                raise ValueError('access_code does not exist and must be defined')
            else:
                access_code = self.default_access_code
        payload = {
            'name': self.name,
            'displayName': self.display_name,
            'defaultAccessCodeEnabled': True,
            'defaultAccessCode': access_code
        }
        try:
            webex_api_call(
                'put',
                f'v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}',
                payload=payload
            )
        except wxcadm.APIError:
            return False
        self.default_access_code_enabled = True
        self.default_access_code = access_code
        return True

    def disable_default_access_code(self) -> bool:
        """ Disable the Default Access Code for handsets

        Returns:
            bool: True on success, False otherwise

        """
        payload = {
            'name': self.name,
            'displayName': self.display_name,
            'defaultAccessCodeEnabled': False,
            'defaultAccessCode': self.default_access_code
        }
        try:
            webex_api_call(
                'put',
                f'v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}',
                payload=payload
            )
        except wxcadm.APIError:
            return False
        self.default_access_code_enabled = False
        return True

    def add_base_stations(self, mac_list: list) -> list[DECTBaseStation]:
        """ Add a Base Station to the DECT Network.

        Args:
            mac_list (list[str]): List of MAC addresses as strings

        Returns:
            list[DECTBaseStation]: List of Base Station entries for the DECT Network, including previously-existing ones

        Raises:
            wxcadm.APIError: Raised when an error is returned by the Webex API

        """
        log.info(f"Adding DECTBaseStations to DECTNetwork {self.id}")
        log.debug(f"MAC List: {mac_list}")
        # Create a DECTBaseStation for now, just to expand later
        payload = []
        for entry in mac_list:
            payload.append(entry)
        try:
            webex_api_call(
                'post',
                f"v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}/baseStations",
                payload={'baseStations': payload})
        except wxcadm.APIError:
            raise wxcadm.APIError
        return self.base_stations

    @property
    def base_stations(self):
        base_stations = []
        response = webex_api_call(
            'get',
            f'v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}/baseStations'
        )
        for item in response['baseStations']:
            this_base = DECTBaseStation(dect_network=self, config=item)
            base_stations.append(this_base)
        return base_stations

    def add_handset(self,
                    display_name: str,
                    line1: Person | Workspace,
                    line2: Optional[Person | Workspace] = None
                    ) -> bool:
        """ Add a handset to the DECT Network.

        Args:
            display_name (str): The text to display on the handset. 16 characters max.
            line1 (Person|Workspace): The :class:`Person` or :class:`Workspace` to assign to Line 1 of the handset
            line2 (Person|Workspace, optional): The Person or Workspace to assign to Line 2, if desired.

        Returns:
            bool: True on success. False otherwise.

        Raises:
            wxcadm.exceptions.APIError: Raised if the Webex API returns an error

        """
        log.info(f"Adding a handset to DECT Network {self.id}")
        payload = {
            'line1MemberId': line1.id,
            'customDisplayName': display_name
        }
        if line2:
            payload['line2MemberId'] = line2.id
        response = webex_api_call(
            'post',
            f"v1/telephony/config/locations/{self.location_id}/dectNetworks/{self.id}/handsets",
            payload=payload)
        return response

    def delete_handset(self, handset: DECTHandset) -> list:
        """ Delete the specified Handset

        Args:
            handset (DECTHandset): The :class:`DECTHandset` instance to delete

        Returns:
            list: The remaining DECT Handsets. An empty list will be returned if no handsets remain

        """
        remaining = []
        for hs in self.handsets:
            if hs.id == handset.id:
                hs.delete()
            else:
                remaining.append(hs)
        return remaining

    def get_base_station(self, mac: str) -> Optional[DECTBaseStation]:
        """ Get the :class:`DECTBaseStation` instance by the Base Station MAC address.

        .. note::

            This currently only works with Base Stations that are being added during the wxcadm session. The Webex API
            does not yet support a method to retrieve existing Base Stations.

        Args:
            mac (str): The MAC address to find

        Returns:
            DECTBaseStation: The DECTBaseStation instance. None is returned if no match is found.

        """
        bs: DECTBaseStation
        for bs in self.base_stations:
            if bs.mac.upper() == mac.upper().strip('-').strip(':'):
                return bs
        return None


class DECTNetworkList(UserList):
    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location]):
        log.info("Initializing empty DECTNetworkList")
        super().__init__()
        self.parent = parent
        self.data = self._get_data()

    def _get_data(self):
        log.info("Getting list of DECT Networks")
        networks = []
        if isinstance(self.parent, wxcadm.Location):
            log.debug(f"Using Location {self.parent.name} as Location filter")
            params = {'locationId': self.parent.id}
        elif isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org {self.parent.name} as Org filter")
            params = {'orgId': self.parent.id}
        else:
            raise ValueError("Invalid parent class")

        response = webex_api_call('get', 'v1/telephony/config/dectNetworks', params=params)
        for network in response['dectNetworks']:
            this_network = DECTNetwork(config=network)
            networks.append(this_network)
        return networks

    def refresh(self) -> DECTNetworkList:
        """ Refresh the list of DECT Networks from Webex

        Returns:
            DECTNetworkList: The refreshed list of DECT Networks

        """
        self.data = self._get_data()
        return self

    def with_base_stations(self, count: Optional[int] = 1) -> list:
        """ Get a list of :class:`DECTNetwork` instances with at least one Base Station

        Args:
            count(int, optional): The minimum number of Base Stations to include in the response. When omitted, any
            Network with at least 1 Base Station will be returned.

        Returns:
            list: A list of :class:`DECTNetwork` instances

        """
        retval = []
        for network in self.data:
            if network.number_of_base_stations >= count:
                retval.append(network)
        return retval

    def with_handsets(self, count: Optional[int] = 1) -> list:
        """

        Args:
            count:

        Returns:

        """

    def create(self,
               name: str,
               model: str,
               default_access_code: Optional[str] = None,
               handset_display_name: Optional[str] = None,
               location: Optional[wxcadm.Location] = None
               ) -> DECTNetwork:
        """ Create a DECT Network

        Currently model DBS110 and DBS210 are supported. You can pass the model as `'DBS110'`, `'110'`, `'DBS210'` or
        `'210'`.

        Args:
            name (str): The name of the DECT Network
            model (str): The model of the base stations that will be in the DECT Network.
            location (Location, optional): This is only optional when the DECTNetworkList is Location-based
            default_access_code (str, optional): The Default Access Code for all handsets on the DECT Network
            handset_display_name (str, optional): The default Display Name for Handsets

        Returns:
            DECTNetwork: The created :class:`DECTNetwork`

        Raises:
            ValueError: Raised when called on an Org-level DECTNetworkList without a provided `location`
            wxcadm.APIError: Raised when an error is returned by the Webex API

        """
        if location is None and not isinstance(self.parent, wxcadm.Location):
            log.warning('DECTNetworkList.create() called without a location for Org-level list')
            raise ValueError('location must be provided for Org-level create')
        if location is not None and not isinstance(location, wxcadm.Location):
            log.warning('DECTNetworkList.create() received bad value for location')
            raise ValueError('location must be of type wxcadm.Location')
        if location is None and isinstance(self.parent, wxcadm.Location):
            location = self.parent

        log.info(f"Creating DECTNetwork {name} in Location: {location.name}")
        # Standardize the model name
        if '110' in model:
            model = 'DMS Cisco DBS110'
            log.debug(f"Normalizing model name to {model}")
        elif '210' in model:
            model = 'DMS Cisco DBS210'
            log.debug(f"Normalizing model name to {model}")
        else:
            log.warning(f"Model {model} not recognized")
            raise ValueError("Model not recognized")
        payload = {
            'name': name,
            'model': model,
            'displayName': handset_display_name
        }
        if default_access_code:
            log.debug(f"Enabling Default Access Code {default_access_code}")
            payload.update({'defaultAccessCodeEnabled': True, 'defaultAccessCode': default_access_code})

        response = webex_api_call('post', f"v1/telephony/config/locations/{location.id}/dectNetworks",
                                  payload=payload)
        new_network = DECTNetwork(location=location, id=response['dectNetworkId'])
        self.data.append(new_network)
        return new_network

    def delete(self, network: DECTNetwork):
        """ Delete a DECT Network

        Args:
            network (DECTNetwork): The DECTNetwork instance of the network to be deleted

        Returns:
            DECTNetworkList: The remaining list of DECT Networks

        """
        self.data = self._get_data()
        remaining_networks = []
        log.debug(f'Looking for DECT Network to delete matching ID {network.id}')
        for item in self.data:
            log.debug(f'Checking ID: {item.id}')
            if item.id == network.id:
                log.debug(f'Found DECT Network: {item.id}. Deleting.')
                success = item.delete()
                if success is False:
                    log.warning('DECT Network delete failed')
                    remaining_networks.append(item)
            else:
                remaining_networks.append(item)
        self.data = remaining_networks
        return self



from __future__ import annotations

from wxcadm import log, Org
from .common import *
from .location import Location


class Wholesale:
    def __init__(self,
                 access_token: str):
        log.info("Wholesale instance initialized")
        # The access token is the only thing that we need to get started
        self._access_token: str = access_token
        # The Authorization header is going to be used by every API call in the package.
        self._headers: dict = {"Authorization": "Bearer " + access_token}
        log.debug(f"Setting Org._headers to {self._headers}")
        log.debug(f"Setting Global _webex_headers")
        global _webex_headers
        _webex_headers['Authorization'] = "Bearer " + access_token
        self._orgs = None

    @property
    def headers(self):
        """The "universal" HTTP headers with the Authorization header present"""
        return self._headers

    @property
    def customers(self):
        response = webex_api_call('get', '/v1/wholesale/customers')
        customer_list = []
        for customer in response:
            this_customer = WholesaleCustomer(self, customer)
            customer_list.append(this_customer)
        return customer_list

    @property
    def orgs(self):
        if self._orgs is None:
            orgs = []
            for customer in self.customers:
                orgs.append(Org(name=customer.external_id, id=customer.id, parent=self))
            self._orgs = orgs
        return self._orgs

    def get_customer(self, id: str = None, name: str = None, spark_id: str = None):
        if id is None and name is None and spark_id is None:
            raise ValueError("A search argument must be provided")
        if id is not None:
            for customer in self.customers:
                if customer.id == id:
                    return customer
        if name is not None:
            for customer in self.customers:
                if customer.external_id == name:
                    return customer
        if spark_id is not None:
            for customer in self.customers:
                if customer.spark_id == spark_id:
                    return customer
        return None


class WholesaleCustomer:
    # Eventually, this should probably be a subclass of Org, since they have the same behaviors and attributes
    def __init__(self, partner, customer_data: dict):
        self.id = customer_data.get('id')
        self.org_id = customer_data.get('orgId')
        self.external_id = customer_data.get('externalId')
        self.address = customer_data.get('address')
        self.status = customer_data.get('status')
        self.packages = customer_data.get('packages')
        self.resource_details = customer_data.get('resourceDetails')

        # Set the Authorization header based on how the instance was built
        self._headers = partner.headers

    @property
    def spark_id(self):
        """ The decoded "Spark ID" of the Org ID"""
        return decode_spark_id(self.id)

    @property
    def locations(self):
        locations = []
        params = {"orgId": self.org_id}
        response = webex_api_call('get', '/v1/locations', params=params)
        for location in response:
            this_location = Location(self, location.get('id'), location.get('name'), time_zone=location.get('timeZone'),
                                     preferred_language=location.get('preferredLanguage'),
                                     announcement_language=location.get('announcementLanguage', None))
            locations.append(this_location)
        return locations

    def get_location(self, id: str = None, name: str = None, spark_id: str = None):
        """ Get the Location instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Locations will be
        searched in order by ID, Name, and finally Spark ID. If no arguments are provided, the method will raise an
        Exception.

        Args:
            id (str, optional): The Location ID to find
            name (str, optional): The Location Name to find
            spark_id (str, optional): The Spark ID to find

        Returns:
            Location: The Location instance correlating to the given search argument.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and spark_id is None:
            raise ValueError("A search argument must be provided")
        if id is not None:
            for location in self.locations:
                if location.id == id:
                    return location
        if name is not None:
            for location in self.locations:
                if location.name == name:
                    return location
        if spark_id is not None:
            for location in self.locations:
                if location.spark_id == spark_id:
                    return location
        return None

    def add_subscriber(self,
                       email: str,
                       package: str,
                       first_name: str,
                       last_name: str,
                       phone_number: str,
                       extension: str,
                       location: Location):
        payload = {
            'customerId': self.id,
            'email': email,
            'package': package,
            'provisioningParameters': {
                'firstName': first_name,
                'lastName': last_name,
                'primaryPhoneNumber': phone_number,
                'extension': extension,
                'locationId': location.id
            }
        }
        response = webex_api_call('post', '/v1/wholesale/subscribers', payload=payload)
        return response




from __future__ import annotations

import humps

import wxcadm.location
from wxcadm import log
from .common import *
from .exceptions import PutError


class RealtimeClass:
    def __init__(self):
        data: dict = webex_api_call('get', self.data_url)
        log.debug(f"[RealTime] Received {data}")
        for key, val in data.items():
            log.debug(f"[RealTime] Setting {key} to {val}")
            self.__setattr__(humps.decamelize(key), val)
            self._api_fields.append(humps.decamelize(key))
        self._initialized = True

    def config_json(self) -> dict:
        return_dict = {}
        for field in self._api_fields:
            value = self.__getattribute__(field)
            log.debug(f"[RealTime] Adding {field} to payload")
            return_dict[humps.camelize(field)] = value
        log.debug(f"[RealTime] Returning {return_dict}")
        return return_dict

    def __setattr__(self, name, value):
        if self._initialized:
            old_val = self.__getattribute__(name)
            log.info(f"[RealTime] Changing {name} from {old_val} to {value}")
            object.__setattr__(self, name, value)
            try:
                webex_api_call('put', self.data_url,
                               payload=self.config_json())
            except wxcadm.exceptions.APIError as e:
                log.warning("[RealTime] The change failed.")
                object.__setattr__(self, name, old_val)
                raise PutError(e)
        else:
            object.__setattr__(self, name, value)


import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice

class TestWorkspaces(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_workspace_list(self):
        workspace_list = self.webex.org.workspaces
        self.assertIsInstance(workspace_list, wxcadm.workspace.WorkspaceList)

    def test_workspace_add_delete(self):
        with self.subTest("Workspace Add"):
            workspace_list: wxcadm.workspace.WorkspaceList = self.webex.org.workspaces
            test_location = choice(self.webex.org.locations.webex_calling())
            new_workspace = workspace_list.create(location=test_location,
                                                  name="wxcadm API Test",
                                                  capacity=1,
                                                  extension="53553",
                                                  notes="Created by wxcadm unit tests")
            self.assertIsInstance(new_workspace, wxcadm.workspace.Workspace)

        with self.subTest("Workspace Delete"):
            new_workspace.delete()
            deleted_workspace = self.webex.org.workspaces.get_by_id(new_workspace.id)
            self.assertEqual(deleted_workspace, None)

    def test_location_workspace_list(self):
        workspace_list = self.random_location.workspaces
        self.assertIsInstance(workspace_list, wxcadm.workspace.WorkspaceList)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from keyword import kwlist
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from wxcadm.person import Person
    from wxcadm.workspace import Workspace

from typing import Optional, Union
from collections import UserList
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config

import wxcadm.exceptions
from .common import *
from wxcadm import log


@dataclass_json
@dataclass
class OrgQueueSettings:
    org: wxcadm.Org
    maintain_queue_position_for_sim_ring: bool = field(metadata=config(field_name="maintainQueuePositionForSimRingEnabled"))
    """ Maintain queue position for Simultaneous routing algorithm """
    agent_unavailable_on_bounce: bool = field(metadata=config(field_name="forceAgentUnavailableOnBouncedEnabled"))
    """ Change the status of a Customer Experience Essentials agent to unavailable in the event of bounced calls """
    play_barge_in_tone: bool = field(metadata=config(field_name="playToneToAgentForBargeInEnabled"))
    """ Play a notification tone to the Agent for Supervisor Barge-In """
    play_monitoring_tone: bool = field(metadata=config(field_name="playToneToAgentForSilentMonitoringEnabled"))
    """ Play a notification tone to the Agent for Supervisor Monitoring """
    play_whisper_tone: bool = field(metadata=config(field_name="playToneToAgentForSupervisorCoachingEnabled"))
    """ Play a notification tone to the Agent for Supervisor Coaching """

    def set(self,
            maintain_queue_position_for_sim_ring: Optional[bool] = None,
            agent_unavailable_on_bounce: Optional[bool] = None,
            play_barge_in_tone: Optional[bool] = None,
            play_monitoring_tone: Optional[bool] = None,
            play_whisper_tone: Optional[bool] = None,
            ):
        """ Set one or more attributes

        Args:
            maintain_queue_position_for_sim_ring (bool, optional): Set this value
            agent_unavailable_on_bounce (bool, optional): Set this value
            play_barge_in_tone (bool, optional): Set this value
            play_monitoring_tone (bool, optional): Set this value
            play_whisper_tone (bool, optional): Set this value

        Returns:
            bool: True on success, False on failure

        """
        if maintain_queue_position_for_sim_ring is not None:
            self.maintain_queue_position_for_sim_ring = maintain_queue_position_for_sim_ring
        if agent_unavailable_on_bounce is not None:
            self.agent_unavailable_on_bounce = agent_unavailable_on_bounce
        if play_barge_in_tone is not None:
            self.play_barge_in_tone = play_barge_in_tone
        if play_monitoring_tone is not None:
            self.play_monitoring_tone = play_monitoring_tone
        if play_whisper_tone is not None:
            self.play_whisper_tone = play_whisper_tone
        payload = self.to_dict()
        log.debug(payload)
        del payload['org']
        webex_api_call('put', "v1/telephony/config/queues/settings", payload=payload,
                       params={'orgId': self.org.org_id})
        return True


class CallQueue:
    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location], id: str, config: Optional[dict] = None):
        """
        .. note::
            CallQueue instances are normally not instantiated manually and are done automatically with the
            :class:`CallQueueList` class initialization.

        """
        self.parent: Union[wxcadm.Org, wxcadm.Location] = parent
        """The parent org of this Call Queue"""
        self.id: str = id
        """The Webex ID of the Call Queue"""
        self.name: str = config.get('name', '')
        """The name of the Call Queue"""
        self.location_id: str = config.get('locationId', '')
        """The Webex ID of the Location associated with this Call Queue"""
        self.phone_number: str = config.get('phoneNumber', '')
        """The DID of the Call Queue"""
        self.extension: str = config.get('extension', '')
        """The extension of the Call Queue"""
        self.enabled: bool = config.get('enabled', False)
        """True if the Call Queue is enabled. False if disabled"""

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.id

    @property
    def spark_id(self) -> str:
        """ The Spark ID for the Call Queue """
        return decode_spark_id(self.id)

    @property
    def config(self) -> dict:
        """ The configuration of this Call Queue instance """
        log.info(f"Getting Queue config for {self.name}")
        response = webex_api_call("get", f"v1/telephony/config/locations/{self.location_id}/queues/{self.id}")
        config = response
        return config

    @property
    def call_forwarding(self):
        """ The Call Forwarding settings for this Call Queue instance """
        # TODO: The rules within Call Forwarding are weird. The rules come back in this call, but they are
        #       different than the /selectiveRules response. It makes sense to aggregate them, but that probably
        #       requires the object->JSON mapping that we need to do for all classes
        response = webex_api_call("get", f"v1/telephony/config/locations/{self.location_id}/queues/{self.id}"
                                         f"/callForwarding")
        forwarding = response
        return forwarding

    @property
    def agents(self) -> list:
        """ The agents assigned to the Call Queue """
        return self.config['agents']

    def add_agent(self, agent: Union[Person, Workspace],
                  weight: Optional[str] = None,
                  skill: Optional[str] = None,
                  joined: bool = True) -> bool:
        """ Add a Person or Workspace as an Agent to the Call Queue

        Args:
            agent (Person|Workspace): The Person or Workspace to add as the agent
            weight (str, optional): The weight of the agent route when using a weighted Queue
            skill (str, optional): The kill level (1-20) of the agent when using a Skill-Based Queue
            joined (bool, optional): Whether the agent will be joined into the Queue as soon as they are added

        Returns:
            bool: True on success, False otherwise

        """
        pass

    def push(self):
        """Push the contents of the CallQueue.config back to Webex

        Returns:
            CallQueue.config: The updated config attribute pulled from Webex after pushing the change

        """
        log.info(f"Pushing Call Queue config to Webex for {self.name}")
        response = webex_api_call('put', f'v1/telephony/config/locations/{self.location_id}/queues/{self.id}',
                                  payload=self.config)

        return response

    # Method aliases
    get_queue_config = config
    get_queue_forwarding = call_forwarding


class CallQueueList(UserList):
    _endpoint = "v1/telephony/config/queues"
    _endpoint_items_key = "queues"
    _item_endpoint = "v1/telephony/config/locations/{location_id}/queues/{item_id}"
    _item_class = CallQueue

    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location]):
        super().__init__()
        log.debug("Initializing CallQueueList")
        self.parent: Union[wxcadm.Org, wxcadm.Location] = parent
        self.data: list = self._get_data()

    def _get_data(self):
        log.debug("_get_data() started")
        params = {}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as Data filter")
            params['orgId'] = self.parent.id
        elif isinstance(self.parent, wxcadm.Location):
            log.debug(f"Using Location ID {self.parent.id} as Data filter")
            params['locationId'] = self.parent.id
        else:
            log.warn("Parent class is not Org or Location, so all items will be returned")
        response = webex_api_call('get', self._endpoint, params=params)
        log.info(f"Found {len(response[self._endpoint_items_key])} items")

        items = []
        for entry in response[self._endpoint_items_key]:
            items.append(self._item_class(parent=self.parent, id=entry['id'], config=entry))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self, id: Optional[str] = None, name: Optional[str] = None, spark_id: Optional[str] = None,
            uuid: Optional[str] = None):
        """ Get the instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Locations will be
        searched in order by ID, Name, UUID, and Spark ID. If no arguments are provided, the method will raise an
        Exception.

        Args:
            id (str, optional): The Call Queue ID to find
            name (str, optional): The Call Queue Name to find. Case-insensitive.
            spark_id (str, optional): The Spark ID to find
            uuid (str, optional): The UUID to find

        Returns:
            CallQueue: The CallQueue instance correlating to the given search argument.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and spark_id is None and uuid is None:
            raise ValueError("A search argument must be provided")
        if id is not None:
            for item in self.data:
                if item.id == id:
                    return item
        if name is not None:
            for item in self.data:
                if item.name.lower() == name.lower():
                    return item
        if uuid is not None:
            for item in self.data:
                if item.spark_id.split('/')[-1].upper() == uuid.upper():
                    return item
        if spark_id is not None:
            for item in self.data:
                if item.spark_id == spark_id:
                    return item
        return None

    def create(self,
               name: str,
               first_name: str,
               last_name: str,
               phone_number: Optional[str],
               extension: Optional[str],
               call_policies: dict,
               queue_settings: dict,
               language: Optional[str] = None,
               time_zone: Optional[str] = None,
               allow_agent_join: Optional[bool] = False,
               allow_did_for_outgoing_calls: Optional[bool] = False,
               location: Optional[wxcadm.Location] = None
               ):
        """ Create a new Call Queue at the Location

        Args:
            name (str): The name of the Call Queue
            first_name (str): The first name of the Call Queue in the directory and for Caller ID
            last_name (str): The last name of the Call Queue in the directory and for Caller ID
            phone_number (str, None): The DID for the Call Queue. None indicates no DID.
            extension (str, None): The extension for the Call Queue. None indicates no extension.
            call_policies (dict): The callPolicies dict for the Call Queue. Because the format of this dict changes
                often, see the documentation at developer.webex.com for values.
            queue_settings (dict): The queueSettings dict for the Call Queue. Because the format of this dict changes
                often, see the documentation at developer.webex.com for values.
            language (str, optional): The language code (e.g. 'en_us') for the Call Queue. Defaults to the
                announcement_language of the Location.
            time_zone (str, optional): The time zone (e.g. 'America/Phoenix') for the Call Queue. Defaults to the
                time_zone of the Location.
            allow_agent_join (bool, optional): Whether to allow agents to join/leave the Call Queue. Default False.
            allow_did_for_outgoing_calls (bool, optional): Whether to allow the Call Queue DID to be used for
                outgoing calls. Default False.
            location (:class:`Location`, optional): The Location to create the Call Queue at. This is required when the
                :class:`CallQueueList` exists at the :class:`Org` level, but not at the :class:`Location` level, because
                the Location is implied at the Location level.

        Returns:
            CallQueue: The created CallQueue instance

        Raises:
            ValueError: Raised when phone_number and extension are both None. One or both is required. Also raised when
                ``location`` is None and the :class:`CallQueueList` is at the Org level.

        """
        # Get some values if they weren't passed
        if phone_number is None and extension is None:
            raise ValueError("phone_number and/or extension are required")
        if location is None and isinstance(self.parent, wxcadm.Org):
            raise ValueError("location is required for Org-level CallQueueList")
        elif location is None and isinstance(self.parent, wxcadm.Location):
            location = self.parent
        log.info(f"Creating Call Queue at Location {location.name} with name: {name}")
        if location.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        if time_zone is None:
            log.debug(f"Using Location time_zone {location.time_zone}")
            time_zone = location.time_zone
        if language is None:
            log.debug(f"Using Location announcement_language {location.announcement_language}")
            language = location.announcement_language

        payload = {
            "name": name,
            "firstName": first_name,
            "lastName": last_name,
            "extension": extension,
            "phoneNumber": phone_number,
            "callPolicies": call_policies,
            "queueSettings": queue_settings,
            "timeZone": time_zone,
            "languageCode": language,
            "allowAgentJoinEnabled": allow_agent_join,
            "phoneNumberForOutgoingCallsEnabled": allow_did_for_outgoing_calls
        }
        response = webex_api_call("post", f"v1/telephony/config/locations/{location.id}/queues",
                                  payload=payload)
        new_queue_id = response['id']

        # Get the details of the new Queue
        webex_api_call("get", f"v1/telephony/config/locations/{location.id}/queues/{new_queue_id}")
        self.refresh()
        return self.get(id=new_queue_id)



from __future__ import annotations

from typing import Optional
from datetime import datetime, timedelta

import wxcadm.exceptions
from .common import *
from wxcadm import log


class LegPart:
    """ A Call Leg is made up of one or two Leg Parts """
    def __init__(self, record: dict):
        log.debug(f"Adding LegPart {record['Local call ID']}")
        self.record = record
        self.start_time = datetime.strptime(record['Start time'], "%Y-%m-%dT%H:%M:%S.%fZ")
        self.end_time = datetime.strptime(record['Release time'], "%Y-%m-%dT%H:%M:%S.%fZ")
        self.answer_time = None
        self.answered = True if record['Answered'] == 'true' else False
        self.local_id = record['Local call ID'] if record['Local call ID'] else None
        self.remote_id = record['Remote call ID'] if record['Remote call ID'] else None
        self.direction: str = record['Direction']
        self.calling_line_id: str = record['Calling line ID']
        self.called_line_id: str = record['Called line ID']
        self.dialed_digits: str = record['Dialed digits']
        self.user_number: str = record['User number']
        self.called_number: str = record['Called number']
        self.calling_number: str = record['Calling number']
        self.redirecting_number: str = record['Redirecting number']
        self.location_name: str = record['Location']
        self.user_type: str = record['User type']
        self.user: str = record['User']
        self.call_type = record['Call type']
        self.duration = record['Duration']
        self.time_offset = record['Site timezone']
        self.releasing_party: str = record['Releasing party']
        self.original_reason: str = record['Original reason']
        self.redirect_reason: str = record['Redirect reason']
        self.related_reason: str = record['Related reason']
        self.outcome: str = record['Call outcome']
        self.outcome_reason: str = record['Call outcome reason']
        self.ring_duration: str = record['Ring duration']
        try:
            self.device_mac: str = record['Device MAC']
        except KeyError:
            self.device_mac: str = record['Device Mac']
        # Transfer Identifiers
        if record['Call transfer time'] and record['Call transfer time'] != 'NA':
            self.transfer_time = datetime.strptime(record['Call transfer time'], "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            self.transfer_time = None

        if record['Transfer related call ID'] and record['Transfer related call ID'] != 'NA':
            self.transfer_related_call_id: str = record['Transfer related call ID']
        else:
            self.transfer_related_call_id = ''

        if record['Answer time'] and record['Answer time'] != 'NA':
            self.answer_time = datetime.strptime(record['Answer time'], "%Y-%m-%dT%H:%M:%S.%fZ")

        self.pstn_inbound: bool = True if record['Call type'].upper() == 'SIP_INBOUND' else False
        self.internal_call: bool = True if record['Call type'].upper() == 'SIP_ENTERPRISE' else False
        self.pstn_outbound = False
        if record['Call type'].upper() in ['SIP NATIONAL', 'SIP_INTERNATIONAL']:
            self.pstn_outbound = True


class CallLeg:
    """ A Call is made up of one ore more Call Legs """
    def __init__(self):
        # app.logger.debug("Adding Call Leg to Call")
        self.parts: list[LegPart] = []

    @property
    def start_time(self) -> datetime:
        """ The Call Leg start date and time

        Returns:
            datetime

        """
        return self.parts[0].start_time

    @property
    def end_time(self) -> datetime:
        """ The Call Leg end date and time

        Returns:
            datetime

        """
        for part in self.parts:
            if part.transfer_time is not None and part.direction.upper() == 'TERMINATING':
                return part.transfer_time
        return self.parts[-1].end_time

    @property
    def pstn_leg(self) -> bool:
        """ Whether this Call Leg is to or from the PSTN (off-net) """
        for part in self.parts:
            if part.pstn_inbound is True or part.pstn_outbound is True:
                return True
        return False

    @property
    def duration(self) -> int:
        """ The number of seconds of duration for the Call Leg """
        return int(self.parts[0].duration)

    @property
    def ring_duration(self) -> int:
        """ The number of seconds that the call was in a ringing state """
        return int(self.parts[0].ring_duration)

    @property
    def transfer_ids(self) -> list:
        """ List of Transfer IDs associated with the Call Leg

        Returns:
            list[str]

        """
        ids = []
        for part in self.parts:
            if part.transfer_related_call_id != '':
                ids.append(part.transfer_related_call_id)
        return ids

    @property
    def orig_part(self) -> Optional[LegPart]:
        """ The :class:`LegPart` for the originating part of the Call Leg"""
        for part in self.parts:
            if part.direction.upper() == 'ORIGINATING':
                return part
        return None

    @property
    def term_part(self) -> Optional[LegPart]:
        """ The :class:`LegPart` for the terminating part of the Call Leg """
        for part in self.parts:
            if part.direction.upper() == 'TERMINATING':
                return part
        return None

    @property
    def is_queue_leg(self) -> bool:
        """ Whether the Call Leg is to a Call Queue """
        if self.term_part is not None:
            if self.term_part.user_type.upper() == 'CALLCENTERPREMIUM':
                return True
        return False

    @property
    def is_aa_leg(self) -> bool:
        """ Whether the Call Leg is to an Auto Attendant """
        if self.term_part is not None:
            if self.term_part.user_type.upper() == 'AUTOMATEDATTENDANTVIDEO':
                return True
        return False

    @property
    def is_agent_leg(self) -> bool:
        """ Whether the Call Leg is a Call Queue leg to an Agent """
        if self.term_part is not None:
            if self.term_part.user_type.upper() == 'USER' and self.term_part.redirect_reason.upper() == 'CALLQUEUE':
                return True
        return False

    @property
    def is_vm_deposit(self) -> bool:
        """ Whether the Call Leg is a Voicemail deposit"""
        if self.orig_part is not None:
            if self.orig_part.called_line_id == 'Voice Portal Voice Messaging Group':
                if self.orig_part.redirect_reason in ['UserBusy', 'NoAnswer', 'Unconditional']:
                    return True
        return False

    @property
    def label(self) -> str:
        orig_part = self.orig_part
        term_part = self.term_part

        if term_part is not None and orig_part is not None:
            if term_part.user == '':
                if orig_part.called_line_id == 'NA':
                    return term_part.user_number
                else:
                    return orig_part.called_line_id
            else:
                return term_part.user
        elif term_part is not None and orig_part is None:
            if term_part.user == '':
                return term_part.user_number
            else:
                return term_part.user

    @property
    def calling_number(self) -> str:
        """ The Calling Part Number for the Call Leg """
        if self.orig_part is None:
            return self.term_part.calling_number
        else:
            return self.orig_part.calling_number

    @property
    def leg_description(self):
        desc = ""
        for part in self.parts:
            if part.user == '':
                endpoint = part.user_type
            else:
                endpoint = part.user
            if part.direction == "ORIGINATING":
                desc = f"From {endpoint}" + desc
            elif part.direction == 'TERMINATING':
                if part.pstn_inbound:
                    desc = desc + "SIP Inbound"
                desc = desc + f" to {endpoint}"
        return desc

    @property
    def new_leg_description(self) -> str:
        leg_from = ''
        leg_to = ''
        to_lock = False
        from_lock = False

        # Orig leg logic
        if self.orig_part is None:
            if self.term_part.call_type == 'SIP_INBOUND':
                leg_from = f'{self.term_part.calling_number}'
                from_lock = True
        else:
            if self.orig_part.user == '':
                if from_lock is False:
                    leg_from = self.orig_part.calling_number
            else:
                if from_lock is False:
                    leg_from = self.orig_part.user
                    from_lock = True
            if self.orig_part.called_line_id == 'NA' or self.orig_part.called_line_id == '':
                if to_lock is False:
                    leg_to = self.orig_part.called_number
            else:
                if to_lock is False:
                    leg_to = self.orig_part.called_line_id
                    to_lock = True

        # Term Leg logic
        if self.term_part is None:
            if self.orig_part.user == '':
                if from_lock is False:
                    leg_from = self.orig_part.calling_number
            else:
                leg_from = self.orig_part.user
                from_lock = True
            if to_lock is False:
                leg_to = f"{self.orig_part.called_number}"
        else:
            if self.term_part.calling_line_id == 'NA' or self.term_part.calling_line_id == '':
                if from_lock is False:
                    leg_from = self.term_part.calling_number
            else:
                if from_lock is False:
                    leg_from = self.term_part.calling_line_id
                    from_lock = True
            if self.term_part.user == '':
                if to_lock is False:
                    leg_to = self.term_part.called_number
            else:
                if to_lock is False:
                    leg_to = self.term_part.user
                    to_lock = True

        description = f"From {leg_from}"
        if self.out_reason:
            description += f' ({self.out_reason})'
        description += f' to {leg_to}'
        if self.in_reason:
            description += f' ({self.in_reason})'
        return description

    @property
    def in_reason(self) -> Optional[str]:
        """ The Reason a Call Leg was received.

        Returns:
            str: A descriptive reason

        """
        if self.term_part is None:
            return None
        if self.term_part.call_type == 'SIP_INBOUND':
            return 'SIP Inbound'
        if self.term_part.redirect_reason == '':
            return None
            # More to do here when we figure out all the reasons a call will come in
        else:
            return self.term_part.redirect_reason

    @property
    def out_reason(self) -> Optional[str]:
        """ The Reason the Call Leg originated.

        This can be useful in finding Call Queue Call-Back legs

        Returns:
            str: A descriptive reason

        """
        if self.orig_part is None:
            return None
        if self.orig_part.related_reason == '':
            return None
        else:
            if self.orig_part.user_type == 'CallCenterPremium' and self.orig_part.related_reason == 'Unrecognised':
                return 'Queue Callback'
            else:
                return self.orig_part.related_reason

    @property
    def out_label(self):
        """ An experimental text description of the Call Leg """
        label = ''
        if self.orig_part is None:
            # Without an Orig Part, we have to get everything we need from the Term Part
            if self.term_part.call_type == 'SIP_INBOUND':
                label += 'SIP Inbound '
            if self.term_part.calling_line_id != '':
                label += f"from {self.term_part.calling_line_id} ({self.term_part.calling_number}) "
        if self.orig_part.call_type == 'SIP_ENTERPRISE':
            if self.orig_part.user == '':
                pass
            else:
                return self.orig_part.user
            return None

    @property
    def parts_id(self) -> list:
        parts_id = []
        for part in self.parts:
            if part.local_id is not None:
                parts_id.append(part.local_id)
            if part.remote_id is not None:
                parts_id.append(part.remote_id)
        return parts_id

    @property
    def answered(self) -> bool:
        """ Whether the Call Leg was answered

        Returns:
            bool

        """
        answered = False
        for part in self.parts:
            if part.answered is True:
                answered = True
        return answered

    @property
    def answered_label(self) -> str:
        """ The string 'Answered' or 'Unanswered'. Useful for labeling in a UI. """
        label = 'Answered' if self.answered is True else 'Unanswered'
        return label

    def add_part(self, record: dict) -> LegPart:
        part = LegPart(record)
        self.parts.append(part)
        return part


class Call:
    def __init__(self, correlation_id: str):
        self.id: str = correlation_id
        self.legs = []
        self.correlation_ids = [correlation_id]
        log.debug("New Call instance created")

    def add_record(self, record: dict):
        log.debug(f"Adding Record to Call {self.id}")
        # Determine if the Local or Remote CallPart IDs have been seen already
        existing_leg = False
        for leg in self.legs:
            if record['Local call ID'] in leg.parts_id or record['Remote call ID'] in leg.parts_id:
                existing_leg = True
                leg.add_part(record)
        if existing_leg is False:
            new_leg = CallLeg()
            self.legs.append(new_leg)
            new_leg.add_part(record)

    @property
    def start_time(self) -> Optional[datetime]:
        """ The Call Start date and time

        Returns:
            datetime

        """
        earliest: Optional[datetime] = None
        for leg in self.legs:
            if earliest is None or leg.start_time < earliest:
                earliest = leg.start_time
        return earliest

    @property
    def end_time(self) -> Optional[datetime]:
        """ The Call End date and time

        Returns:
            datetime

        """
        latest: Optional[datetime] = None
        for leg in self.legs:
            if latest is None or leg.end_time > latest:
                latest = leg.end_time
        return latest

    @property
    def duration(self):
        """ The duration of the Call

        Returns:
            timedelta

        """
        difference: timedelta = self.end_time - self.start_time
        return str(difference).split('.')[0]

    @property
    def calling_number(self) -> str:
        """ The original Calling Number from the first Leg of the Call

        Returns:
            str: The Calling Number

        """
        return self.legs_sorted[0].calling_number

    @property
    def leg_count(self) -> int:
        """ The number of Legs that make up the call

        Returns:
            int: Number of Legs

        """
        return len(self.legs)

    @property
    def legs_sorted(self):
        """ List of Call Legs sorted by their Start DateTime

        Returns:
            list[CallLeg]: The list of CallLegs

        """
        return sorted(self.legs, key=lambda x: x.start_time, reverse=False)

    @property
    def part_ids(self):
        """ List of Part IDs used within the Call """
        parts = []
        for leg in self.legs:
            parts.extend(leg.parts_id)
        return set(parts)

    @property
    def transfer_ids(self) -> list:
        """ List of Transfer IDs used within the Call """
        ids = []
        for leg in self.legs:
            ids.extend(leg.transfer_ids)
        return ids

    @property
    def _missing_part_ids(self) -> list:
        missing = []
        for id in self.transfer_ids:
            if id not in self.part_ids:
                missing.append(id)
        return missing

    @property
    def answered_legs(self):
        """ The number of Call Legs where the Answer indicator is True

        Returns:
            int: The number of Answered legs

        """
        answered = 0
        for leg in self.legs:
            if leg.answered is True:
                answered += 1
        return answered

    @property
    def has_queue_leg(self) -> bool:
        """ Whether the Call has a Call Leg that involves a Call Queue

        Returns:
            bool

        """
        for leg in self.legs:
            if leg.is_queue_leg is True:
                return True
        return False

    @property
    def has_pstn_leg(self) -> bool:
        """ Whether the Call has a Call Leg that was to or from the PSTN (off-net)

        Returns:
            bool

        """
        for leg in self.legs:
            if leg.pstn_leg:
                return True
        return False

    @property
    def has_vm_deposit_leg(self) -> bool:
        """ Whether the Call contains a Voicemail deposit

        Returns:
            bool

        """
        for leg in self.legs:
            if leg.is_vm_deposit:
                return True
        return False

    @property
    def is_queue_abandon(self) -> bool:
        """ Whether the Call is one where the caller ended the call while waiting in a Call Queue

        Returns:
             bool

        """
        if self.has_queue_leg:
            for leg in self.legs:
                if leg.is_agent_leg and leg.answered:
                    return False
            return True
        else:
            return False


class CallDetailRecords:
    def __init__(self, records: list, webex: Optional[wxcadm.Webex] = None):
        self.records: list = records
        """ The records that were sent to the CallDetailRecords instance """
        self.webex = webex
        """ The wxcadm.Webex connection to use to look up identifiers """
        self._retry_records = []
        self.calls: list[Call] = []
        """ The (unordered) list of Call instances after processing """
        self.__process_calls()
        # Retry any that didn't make it through the first time
        self._to_merge = []
        self.__merge_transfer_calls()

    def get_call_by_correlation_id(self, correlation_id: str) -> Optional[Call]:
        """ Find a Call by its Correlation ID

        Args:
            correlation_id (str): The Correlation ID to search for

        Returns:
            Call: The matching Call instance

        """
        log.debug(f"Getting Call by Correlation ID: {correlation_id}")
        for call in self.calls:
            if correlation_id == call.id:
                return call
        return None

    def get_call_by_part_call_id(self, part_call_id: str):
        """ Find a Call by the Part ID of one of the Call Parts

        Searching by Part iD is rare and normally only used for internal processing, but users who have an idea of
        what the Part ID is might have use for this method.

        Returns:
            Call: The matching Call instance

        """
        log.debug(f"Getting Call by Park Call ID: {part_call_id}")
        for call in self.calls:
            if part_call_id in call.part_ids:
                return call
        return None

    @property
    def calls_sorted(self):
        """ A list of Calls sorted by timestamp """
        return sorted(self.calls, key=lambda x: x.start_time, reverse=False)

    def __user_finder(self, type: str, location_id: str, user_id: str) -> str:
        log.info(f"Finding User with type '{type}' and ID {user_id}")
        if type == 'AutomatedAttendantVideo':
            me = self.webex.org.auto_attendants.get(uuid=user_id)
            if me is not None:
                log.debug(f"Found match: {me.name}")
                return me.name
            else:
                log.warning("No User Match found")
                return 'Auto Attendant'
        elif type == 'CallCenterPremium':
            me = self.webex.org.call_queues.get(uuid=user_id)
            if me is not None:
                log.debug(f"Found match: {me.name}")
                return me.name
            else:
                log.warning("No User Match found")
                return 'Call Queue'
        elif type == 'Place':
            me = self.webex.org.workspaces.get(uuid=user_id)
            if me is not None:
                log.debug(f"Found match: {me.name}")
                return me.name
            else:
                log.warning("No User Match found")
                return 'Workspace'
        else:
            log.warning(f"No Method to find User for {type}")
        return f'{type} - {user_id}'

    def __process_calls(self, records: Optional[list] = None):
        log.debug("Processing calls for CallDetailRecords")
        if records is None:
            records = self.records

        for record in records:
            # app.logger.debug(f"Correlation ID: {record['Correlation ID']}")
            if record['User'] == '' or record['User'] == 'NA':
                record['User'] = self.__user_finder(
                    record['User type'],
                    record['Site UUID'],
                    record['User UUID']
                )
            call = self.get_call_by_correlation_id(record['Correlation ID'])
            # Create a new Call record if we haven't seen this Correlation ID yet
            if call is None:
                log.debug(f"New Correlation ID {record['Correlation ID']}. Creating Call")
                call = Call(record['Correlation ID'])
                self.calls.append(call)
            call.add_record(record)

    def __delete_call(self, id: str):
        new_calls = []
        for call in self.calls:
            if call.id != id:
                new_calls.append(call)
        self.calls = new_calls

    def __merge_transfer_calls(self):
        merge_done = False
        # This is done as a while loop that never ends without a break. Since matching calls will both have each others
        # part_id values, we have to take them one at a time without looking at the self.calls for the loop.
        while merge_done is False:
            to_process = []
            for call in self.calls:
                if len(call._missing_part_ids) > 0:
                    to_process.append(call)

            if len(to_process) == 0:
                break

            call = to_process[0]
            missing_id = call._missing_part_ids[0]

            id_to_find = missing_id
            for search_call in self.calls:
                if missing_id in search_call.part_ids and \
                        (search_call.start_time - timedelta(hours=24) <=
                         call.start_time <=
                         search_call.start_time + timedelta(hours=24)):
                    self.__merge_calls(call, search_call)
                    break

    def __merge_calls(self, call1: Call, call2: Call):
        new_calls_list = []
        for call in self.calls:
            if call == call1:
                call.legs.extend(call2.legs)
                call.correlation_ids.extend(call2.correlation_ids)
                new_calls_list.append(call)
            elif call == call2:
                continue
            else:
                new_calls_list.append(call)
        self.calls = new_calls_list
        return self.calls

    def get_abandoned_calls(self) -> list:
        """ Get a list of Calls where the caller hung up in a Call Queue prior to an Agent answering

        Returns:
            list[Call]: A list of Call instances

        """
        response = []
        for call in self.calls_sorted:
            if call.has_queue_leg and call.is_queue_abandon:
                response.append(call)
        return response

    def get_pstn_calls(self) -> list:
        """ Get a list of Calls where a PSTN leg is present

        Returns:
            list[Call]: A list of Call instances

        """
        response = []
        for call in self.calls_sorted:
            if call.has_pstn_leg:
                response.append(call)
        return response

    def get_voicemail_deposit_calls(self) -> list:
        """ Get a list of Calls where the call was sent to Voicemail for the caller to leave a message

        Returns:
            list[Call]: A list of Call instances

        """
        response = []
        for call in self.calls_sorted:
            if call.has_vm_deposit_leg:
                response.append(call)
        return response



from __future__ import annotations

import os
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    pass
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from collections import UserList
from requests_toolbelt import MultipartEncoder

import wxcadm.org
from wxcadm import log
from .common import *


class AnnouncementList(UserList):
    def __init__(self, parent: wxcadm.Org):
        super().__init__()
        log.debug("Initializing AnnouncementList instance")
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_announcements()

    def _get_announcements(self):
        log.debug("Getting announcements")
        annc_list = []
        announcements = webex_api_call("get", "v1/telephony/config/announcements", params={'locationId': 'all'})
        for annc in announcements.get('announcements', []):
            this_annc = Announcement(parent=self.parent, **annc)
            annc_list.append(this_annc)
        return annc_list

    def _refresh_announcements(self):
        log.debug("Refreshing announcements")
        self.data = self._get_announcements()

    def get(self, id:str = None,
            name:str = None,
            file_name: str = None,
            level: str = None,
            location: wxcadm.Location = None) -> Announcement | list[Announcement] | None:
        """ Get the Announcements matching the given criteria.

        Args:
            id (str, optional): The ID of the Announcement
            name (str, optional): The name of the Announcement
            file_name (str, optional) : The name of the file associated with the Announcement
            level (str, optional) : The level of the Announcement (e.g. ORGANIZATION, LOCATION)
            location (Location, optional) : The Location instance to filter on

        Returns:
            Announcement | list[Announcement] | None: The matching Announcements. If more than one Announcement is found,
                a list of Announcements is returned. If no Announcements are found, None is returned.

        """
        filtered_announcements = []
        for annc in self.data:
            match = True
            if id is not None and annc.id != id:
                match = False
            if name is not None and name not in annc.name:
                match = False
            if file_name is not None and annc.fileName != file_name:
                match = False
            if level is not None and annc.level != level:
                match = False
            if location is not None:
                if annc.level == 'ORGANIZATION':
                    match = False
                else:
                    if annc.location is not None and annc.location['id'] != location.id:
                        match = False

            if match is True:
                filtered_announcements.append(annc)

        if len(filtered_announcements) == 1:
            return filtered_announcements[0]
        elif len(filtered_announcements) > 1:
            return filtered_announcements
        else:
            return None

    def get_by_location_id(self, location_id: str) -> list[Announcement]:
        """ Get the :py:class:`Announcement` instances for the given location_id

        This will only return :py:class:`Announcement` entries which are defined at the Location level. It will not
        include any :py:class:`Announcement` that is defined at the Organization level.

        Args:
            location_id (str): The Location ID to filer on

        Returns:
            list[Announcement]: A list of the :py:class:`Announcement` instances for the given location. If none
                are found, an empty list will be returned.

        """
        log.debug(f"Getting announcements by location id: {location_id}")
        annc_list = []
        for annc in self.data:
            if annc.level == "LOCATION" and annc.location['id'] == location_id:
                annc_list.append(annc)

        return annc_list

    def get_by_id(self, id: str) -> Announcement | None:
        """ Ge the :py:class:`Announcement` instance with the given ID

        Args:
            id (str): The Announcement ID

        Returns:
            Announcement: The matching :py:class:`Announcement` instance. None is returned if no match can be found.

        """
        log.debug(f"Getting announcements by id: {id}")
        for annc in self.data:
            if annc.id == id:
                return annc
        return None

    @property
    def stats(self):
        """ The repository usage for announcements within the Org """
        log.debug("Getting stats")
        response = webex_api_call("get", "v1/telephony/config/announcements/usage", params={"orgId": self.parent.id})
        return response

    def upload(self, name: str, filename: str, location: str | wxcadm.Location = None):
        """ Upload a new announcement file

        Args:
            name (str): The unique name of the Announcement
            filename (str): The path to the WAV file to upload
            location (str | Location, optional): The Location ID or Location instance to associate the uploaded
                announcement to. If not present, the announcement will be uploaded at the Organization level.

        Returns:
            str: The ID of the created Announcement

        """
        log.info(f"Uploading Announcement {name} with file: {filename}")
        upload_as = os.path.basename(filename)
        content = open(filename, "rb")
        must_close = True
        encoder = MultipartEncoder(
            fields={"name": name,
                    "file": (upload_as, content, 'audio/wav')
                    }
        )
        log.debug(f"Location is: {type(location)}")

        if location is not None:
            # Upload at the Location level
            if isinstance(location, wxcadm.location.Location):
                log.info(f"Uploading to Location: {location.name}")
                location_id = location.id
            else:
                log.info(f"Uploading to Location ID: {location}")
                location_id = location
            url = f"v1/telephony/config/locations/{location_id}/announcements"
        else:
            # Upload at the Org level
            log.info("Uploading to Organization")
            url = "v1/telephony/config/announcements"

        response = webex_api_call("post_upload",
                                  url,
                                  payload=encoder)

        if must_close is True:
            content.close()

        log.debug("Response:")
        log.debug(f"\t{response}")
        self._refresh_announcements()
        return response['id']


@dataclass
class Announcement:
    parent: wxcadm.Org = field(repr=False)
    id: str
    name: str
    fileName: str
    fileSize: str = field(repr=False)
    mediaFileType: str = field(repr=False)
    lastUpdated: str = field(repr=False)
    level: str
    location: dict | None = field(default=None)

    @property
    def used_by(self):
        """ The list of call features this announcement is assigned to """
        used_by = []

        if self.level.lower() == "organization":
            url = f"v1/telephony/config/announcements/{self.id}"
        elif self.level.lower() == "location":
            url = f"v1/telephony/config/locations/{self.location['id']}/announcements/{self.id}"
        else:
            raise ValueError("Cannot determine Announcement level")

        response = webex_api_call("get", url)
        if response.get('featureReferences', None) is not None:
            for usage in response.get('featureReferences'):
                item = {
                    "id": usage['id'],
                    "name": usage['name'],
                    "type": usage['type'],
                    "location": self.parent.locations.get(id=usage['locationId']),
                }
                if usage['type'] == 'Call Queue':
                    cq = self.parent.get_call_queue_by_id(usage['id'])
                    item['instance'] = cq
                used_by.append(item)
        return used_by

    @property
    def in_use(self):
        """ Boolean whether the Announcement is being used by a calling feature """
        if len(self.used_by) > 0:
            return True
        else:
            return False

    def replace_file(self, filename: str) -> bool:
        """ Replace the existing audio file with a new one

        Args:
            filename (str): The name of the file to replace the existing audio file with

        Returns:
            bool: True if the file was successfully replaced, False otherwise

        """
        log.info(f"Replacing {self.name} file with file: {filename}")
        upload_as = os.path.basename(filename)
        content = open(filename, "rb")
        must_close = True
        encoder = MultipartEncoder(
            fields={"name": self.name,
                    "file": (upload_as, content, 'audio/wav')
                    }
        )

        if self.level.lower() == "organization":
            log.debug("Replacing at the Organization level")
            url = f"v1/telephony/config/announcements/{self.id}"
        elif self.level.lower() == "location":
            log.debug("Replacing at the Location level")
            url = f"v1/telephony/config/locations/{self.location['id']}/announcements/{self.id}"
        else:
            raise ValueError("Cannot determine Announcement level")

        response = webex_api_call("put_upload", url, payload=encoder)

        if must_close is True:
            content.close()

        log.debug("Response:")
        log.debug(f"\t{response}")
        return True

    def delete(self) -> bool:
        """ Delete the Announcement

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Deleting Announcement {self.name} at the {self.level.title()} level")
        if self.in_use is True:     # No use even trying to delete one in use
            return False

        if self.level.lower() == "organization":
            url = f"v1/telephony/config/announcements/{self.id}"
        elif self.level.lower() == "location":
            url = f"v1/telephony/config/locations/{self.location['id']}/announcements/{self.id}"
        else:
            raise ValueError("Cannot determine Announcement level")

        response = webex_api_call("delete", url)
        log.debug(f"Response: {response}")
        return True


@dataclass_json
@dataclass
class Playlist:
    org: wxcadm.Org = field(repr=False)
    id: str
    """ The ID of the Playlist """
    name: str
    """ The name of the Playlist """
    file_count: int = field(metadata=config(field_name="fileCount"))
    """ The number of files in the Playlist """
    last_updated: str = field(metadata=config(field_name="lastUpdated"))
    """ The timestamp the Playlist was last updated """
    in_use: bool = field(default=False, metadata=config(field_name="isInUse"))
    """ Whether the Playlist is in use anywhere """
    location_count: int = field(default=0, metadata=config(field_name="locationCount"))
    """ The number of Locations the Playlist is associated with """

    # Attributes that will be retrieved later with a __getattr__
    file_size: int = field(init=False, repr=False, metadata=config(field_name="fileSize"))
    """ The total size of the files in the Playlist in bytes """
    announcements: list = field(init=False, repr=False)
    """ List of :class:`Announcement` instances in the Playlist """

    def __getattr__(self, item):
        # The following is a crazy fix for a PyCharm debugger bug. It serves no purpose other than to stop extra API
        # calls when developing in PyCharm. See the following bug:
        # https://youtrack.jetbrains.com/issue/PY-48306
        if item == 'shape':
            return None
        log.debug(f"Collecting details for Playlist: {self.id}")
        response = webex_api_call('get', f"v1/telephony/config/announcements/playlists/{self.id}",
                                  params={'orgId': self.org.id})
        self.file_size = response.get('fileSize', 0)
        self.announcements = []
        for announcement in response.get('announcements', []):
            self.announcements.append(Announcement(parent=self.org, **announcement))
        return self.__getattribute__(item)

    def refresh(self) -> bool:
        """ Refresh the Playlist details from the API.

        Returns:
            bool: True on success, False otherwise

        """
        response = webex_api_call('get', f"v1/telephony/config/announcements/playlists/{self.id}",
                                  params={'orgId': self.org.id})
        self.file_size = response.get('fileSize', 0)
        self.file_count = response.get('fileCount', 0)
        self.last_updated = response.get('lastUpdated', '')
        self.in_use = response.get('isInUse', False)
        self.location_count = response.get('locationCount', 0)
        self.announcements = []
        for announcement in response.get('announcements', []):
            self.announcements.append(Announcement(parent=self.org, **announcement))
        return True

    def delete(self) -> bool:
        """ Delete the Playlist

        Returns:
            bool: True on success, False otherwise

        """
        webex_api_call('delete', f"v1/telephony/config/announcements/playlists/{self.id}",
                       params={'orgId': self.org.id})
        return True

    def replace_announcements(self, announcements: list[Announcement]) -> bool:
        """ Replace the announcements in the Playlist with the given list of Announcements.

        Args:
            announcements (list[Announcement]): The list of Announcements to replace the announcements in the
                Playlist with.

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'announcementIds': []}
        for announcement in announcements:
            payload['announcementIds'].append(announcement.id)
        webex_api_call('put', f"v1/telephony/config/announcements/playlists/{self.id}",
                       params={'orgId': self.org.id},
                       payload=payload)
        self.refresh()
        return True

    @property
    def locations(self) -> list[wxcadm.Location]:
        """ The list of Locations the Playlist is associated with """
        log.debug(f"Getting locations for Playlist: {self.id}")
        response = webex_api_call('get', f"v1/telephony/config/announcements/playlists/{self.id}/locations",
                                  params={'orgId': self.org.id})
        locations = []
        for location in response['locations']:
            if location['playlistId'] == self.id:
                locations.append(wxcadm.Location(parent=self.org, **location))
        return locations

    def assign_to_location(self, location: wxcadm.Location) -> bool:
        """ Assign the Playlist to the given Location

        Args:
             location (Location): The Location to assign the Playlist to

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'locationIds': []}
        current_locations = self.locations
        for current_loc in current_locations:
            payload['locationIds'].append(current_loc.id)
        payload['locationIds'].append(location.id)

        webex_api_call('put', f"v1/telephony/config/announcements/playlists/{self.id}",
                       params={'orgId': self.org.id},
                       payload=payload)
        self.refresh()
        return True


class PlaylistList(UserList):
    def __init__(self, parent: wxcadm.Org):
        super().__init__()
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        response = webex_api_call('get', f"v1/telephony/config/announcements/playlists",
                                  params={'orgId': self.parent.id})
        log.debug(f"Response: {response}")
        data = []
        for item in response['playlists']:
            item['org'] = self.parent
            data.append(Playlist.from_dict(item))
        return data

    def get(self, id: str = '', name: str ='') -> Playlist | None:
        """ Get a :class:`Playlist` by ID or Name

        Args:
            id (str, optional): The ID of the Playlist
            name (str, optional): The Name of the Playlist

        Returns:
            Playlist: The matching :class:`Playlist`
            None: If no Playlist was found

        """
        for playlist in self.data:
            if id != '' and playlist.id == id:
                return playlist
            if name != '' and playlist.name == name:
                return playlist
        return None

    def create(self, name: str, announcements: list[Announcement]) -> Playlist:
        """ Create a new Playlist

        Args:
            name (str): The name of the Playlist
            announcements (list[Announcement]): A list of :class:`Announcement` instances to associate with the Playlist

        Returns:
            Playlist: The created Playlist

        """
        log.info(f"Creating Playlist: {name}")
        payload = {'name': name,
                   'announcementIds': []}
        for announcement in announcements:
            payload['announcementIds'].append(announcement.id)
        response = webex_api_call('post', f"v1/telephony/config/announcements/playlists",
                              payload=payload,
                              params={'orgId': self.parent.id})
        new_playlist_info = webex_api_call('get', f"v1/telephony/config/announcements/playlists/{response['id']}",
                                           params={'orgId': self.parent.id})
        new_playlist_info['org'] = self.parent
        log.debug(f"Response: {new_playlist_info}")
        this_playlist = Playlist.from_dict(new_playlist_info)
        self.data.append(this_playlist)
        return this_playlist





from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union
from collections import UserList
import wxcadm
from .common import *
from .exceptions import *
from wxcadm import log
if TYPE_CHECKING:
    from wxcadm import Org, Location


class VirtualLine:
    def __init__(self, parent: Union[Org, Location], config: Optional[dict] = None):
        self.parent = parent
        """ The :class:`Org` or :class:`Location` to which the Virtual Line belongs """
        self.id: str = ''
        """ The ID of the Virtual Line """
        self.first_name: str = ''
        """ The first name of the Virtual Line """
        self.last_name: str = ''
        """ The last name of the Vierual Line """
        self.caller_id_first_name: str = ''
        """ The Caller ID first name of the Virtual Line """
        self.caller_id_last_name: str = ''
        """ The Caller ID last name of the Virtual Line """
        self.caller_id_number: str = ''
        """ The Caller ID number of the Virtual Line """
        self.external_caller_id_policy: str = ''
        """ The Caller ID Policy for presenting Caller ID externally """
        self.custom_external_caller_id_name: str = ''
        """ The custom Caller ID Name used for external calls """
        self.extension: Optional[str] = None
        """ The Virtual Line extension """
        self.esn: Optional[str] = None
        """ The Virtual Line ESN """
        self.phone_number: Optional[str] = None
        """ The Virtual Line phone number """
        self.location_id: str = ''
        """ The Location ID of the Virtual Line """
        self.billing_plan: str = ''
        """ The Virtual Line Billing Plan """
        self.num_devices_assigned: int = 0
        """ The number of devices that are assigned to the Virtual Line """

        # Dynamic attributes pulled by properties
        self._display_name: Optional[str] = None
        self._directory_search_enabled: Optional[bool] = None
        self._announcement_language: Optional[str] = None
        self._time_zone: Optional[str] = None
        self._devices: Optional[list] = None

        if config:
            self._process_config(config)

    def _process_config(self, config: dict) -> bool:
        self.id = config.get('id', '')
        self.first_name = config.get('firstName', '')
        self.last_name = config.get('lastName', '')
        self.caller_id_first_name = config.get('callerIdFirstName', '')
        self.caller_id_last_name = config.get('callerIdLastName', '')
        self.caller_id_number = config.get('callerIdNumber', '')
        self.external_caller_id_policy = config.get('externalCallerIdNamePolicy', '')
        if 'number' in config.keys():
            self.extension: Optional[str] = config['number'].get('extension', None)
            self.esn: Optional[str] = config['number'].get('esn', None)
            self.phone_number: Optional[str] = config['number'].get('external', None)
        if 'location' in config.keys():
            self.location_id: str = config['location'].get('id', '')
        self.billing_plan: str = config.get('billingPlan', '')
        self.num_devices_assigned: int = config.get('numberOfDevicesAssigned', 0)

        return True

    def _get_details(self) -> bool:
        config = webex_api_call('get', f'v1/telephony/config/virtualLines/{self.id}')
        self._display_name = config.get('displayName', '')
        self._directory_search_enabled = config.get('directorySearchEnabled', False)
        self._announcement_language = config.get('announcementLanguage', '')
        self._time_zone = config.get('timeZone', '')
        return True

    @property
    def org_id(self) -> str:
        """ The Org ID of the Virtual Line """
        return self.parent.org_id

    @property
    def display_name(self) -> str:
        """ The Display Name of the Virtual Line"""
        if self._display_name is None:
            self._get_details()
        return self._display_name

    @property
    def directory_search_enabled(self) -> bool:
        """ Flag to indicate Directory Search """
        if self._directory_search_enabled is None:
            self._get_details()
        return self._directory_search_enabled

    @property
    def announcement_language(self) -> str:
        """ The announcement language """
        if self._announcement_language is None:
            self._get_details()
        return self._announcement_language

    @property
    def time_zone(self) -> str:
        """ The time zone of the Virtual Line"""
        if self._time_zone is None:
            self._get_details()
        return self._time_zone

    def get_monitored_by(self):
        """ Returns a list of Users (Person) and Workspaces that are Monitoring this VirtualLine """
        if isinstance(self.parent, wxcadm.Org):
            monitor_list = self.parent.get_all_monitoring()
        elif isinstance(self.parent, wxcadm.Location):
            monitor_list = self.parent.parent.get_all_monitoring()
        try:
            return monitor_list['virtual_lines'][self.id]
        except (KeyError, TypeError):
            return None

    def get_call_recording(self):
        """The Call Recording config for the Person

        Returns:
            dict: The Call Recording config for the Person instance

        """
        log.info(f"Getting Call Recording config for {self.display_name}")
        response = webex_api_call('get',
                                             f"v1/telephony/config/virtualLines/{self.id}/callRecording")
        return response

    def enable_call_recording(self, type: str,
                              record_vm: bool = False,
                              announcement_enabled: bool = False,
                              reminder_tone: bool = False,
                              reminder_interval: int = 30,
                              can_play: bool = True,
                              can_download: bool = True,
                              can_delete: bool = True,
                              can_share: bool = True,
                              transcribe: bool = True,
                              ai_summary: bool = True):
        """ Enable and configure Call Recording for the Virtual Line

        .. note::

            Some parameters, such as ``'transcribe'`` and ``'can_play'`` only apply to call recording done by the Webex
            platform. If the recording is being done by another provider, these parameters will have no effect.

        Args:
            type (str): The type of Call Recording. Value must be 'always', 'never', 'always_with_pause' or 'on_demand'
            record_vm (bool, optional): Whether to record Voicemail. Defaults to False
            announcement_enabled (bool, optional): Whether to announce Call Recording. Defaults to False
            reminder_tone (bool, optional): Whether to play a reminder tone. Defaults to False
            reminder_interval (int, optional): Interval in seconds between reminders. Defaults to 30
            can_play (bool, optional): Whether the user can play recordings. Defaults to True.
            can_download (bool, optional): Whether the user can download recordings. Defaults to True.
            can_delete (bool, optional): Whether the user can delete recordings. Defaults to True.
            can_share (bool, optional): Whether the user can share recordings. Defaults to True.
            transcribe (bool, optional): Enable AI transcription of recordings. Defaults to True.
            ai_summary (bool, optional): Enable AI summary of recordings. Defaults to True.

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Enabling Call Recording for {self.display_name}")
        type_map = {"always": "Always",
                    "never": "Never",
                    "always_with_pause": "Always with Pause/Resume",
                    "on_demand": "On Demand with User Initiated Start"}
        if type not in type_map.keys():
            raise ValueError("'type' must be 'always', 'never', 'always_with_pause' or 'on_demand'.")
        payload = {"enabled": True,
                   "record": type_map[type],
                   "recordVoicemailEnabled": record_vm,
                   "startStopAnnouncementEnabled": announcement_enabled,
                   "notification": {
                       "type": "Beep",
                       "enabled": reminder_tone,
                       "repeat": {
                           "enabled": reminder_tone,
                           "interval": reminder_interval
                       }
                   },
                   "callRecordingAccessSettings": {
                       "viewAndPlayRecordingsEnabled": can_play,
                       "downloadRecordingsEnabled": can_download,
                       "deleteRecordingsEnabled": can_delete,
                       "shareRecordingsEnabled": can_share
                   },
                   "postCallRecordingSettings": {
                       "transcriptEnabled": transcribe,
                       "summaryAndActionItemsEnabled": ai_summary
                   }
                   }
        response = webex_api_call('put', f"v1/telephony/config/virtualLines/{self.id}/callRecording",
                                  payload=payload, params={'orgId': self.org_id})
        log.debug(f"Response: {response}")
        return True

    def disable_call_recording(self):
        """ Disables Call Recording for the Virtual Line

        This method will return True even if the Virtual Line did not have Call Recording enabled prior to calling
        the method.

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Disabling Call Recording for {self.display_name}")
        recording_config = {'enabled': False}
        response = webex_api_call('put', f"v1/telephony/config/virtualLines/{self.id}/callRecording",
                                  payload=recording_config, params={'orgId': self.org_id})
        log.debug(f"Response: {response}")
        return True

    def refresh_config(self) -> bool:
        """ Refresh the Virtual Line configuration from Webex This is especially useful when a new Virtual Line is
        created and the configuration is not known. """
        log.info(f"Refreshing Virtual Line config: {self.id}")
        response = webex_api_call('get', f'v1/telephony/config/virtualLines', params={'id': self.id})
        self._process_config(response['virtualLines'][0])
        return True

    def delete(self) -> bool:
        """ Delete the Virtual Line """
        log.info(f"Deleting Virtual Line: {self.first_name} {self.last_name} ({self.id})")
        webex_api_call('delete', f'v1/telephony/config/virtualLines/{self.id}')
        return False

    def _build_put_payload(self) -> dict:
        payload = {
            'firstName': self.first_name,
            'lastName': self.last_name,
            'displayName': self.display_name,
            'phoneNumber': self.phone_number,
            'extension': self.extension,
            'announcementLanguage': self.announcement_language,
            'callerIdFirstName': self.caller_id_first_name,
            'callerIdLastName': self.caller_id_last_name,
            'callerIdNumber': self.caller_id_number,
            'timeZone': self.time_zone
        }
        return payload

    def update(self, **kwargs):
        """ Update the Virtual Line

        To update attributes, pass those attributes as arguments. For example, to update the :attr:`phone_number`, call
        the method as `update(phone_number='+15552345678')`. To update multiple attributes, include them as arguments.
        For example, `update(first_name='Joe', last_name='User')` will update both the first and last name.

        Args:
            **kwargs: All attributes are supported except :attr:`devices`, wghich has its own methods

        Returns:
            bool: True on success, False otherwise

        Raises:
            AttributeError: Raised when attempting to change a non-existing attribute
            APIError: Raised when the Webex API rejects the change

        """
        log.info(f"Updating Virtual Line with ID {self.id}")
        for attr, val in kwargs.items():
            getattr(self, attr)
            setattr(self, attr, val)
        response = webex_api_call('put', f'v1/telephony/config/virtualLines/{self.id}',
                                  payload=self._build_put_payload())
        return True

    @property
    def ecbn(self) -> dict:
        """ The Emergency Callback Number details of the Virtual Line """
        response = webex_api_call('get', f'v1/telephony/config/virtualLines/{self.id}/emergencyCallbackNumber',
                                  params={'orgId': self.parent.org_id})
        return response

    def set_ecbn(self, value: Union[str, wxcadm.Person, wxcadm.Workspace, wxcadm.VirtualLine]):
        """ Set the ECBN of the Virtual Line

        Valid values are ``'direct'``, ``'location'``, or a :class:`Person`, :class:`Workspace`, or
        :class:`VirtualLine` to set the ECBN to one of those.

        Args:
            value (str, Person, Workspace, VirtualLine): The value to set the ECBN to

        Returns:
            bool: True on success

        """
        if isinstance(value, wxcadm.Person) or \
                isinstance(value, wxcadm.Workspace) or \
                isinstance(value, wxcadm.VirtualLine):
            payload = {
                'selected': 'LOCATION_MEMBER_NUMBER',
                'locationMemberId': value.id
            }
        elif value.lower() == 'direct' or value.lower() == 'direct_line':
            payload = {'selected': 'DIRECT_LINE'}
        elif value.lower() == 'location' or value.lower() == 'location_ecbn':
            payload = {'selected': 'LOCATION_ECBN'}
        else:
            raise ValueError('Unknown value')

        response = webex_api_call('put', f'v1/telephony/config/virtualLines/{self.id}/emergencyCallbackNumber',
                                  params={'orgId': self.org_id}, payload=payload)
        return response


class VirtualLineList(UserList):
    _endpoint = "v1/telephony/config/virtualLines"
    _endpoint_items_key = 'virtualLines'
    _item_endpoint = "v1/telephony/config/virtualLines/{item_id}"
    _item_class = VirtualLine

    def __init__(self, parent: Union[Org, Location]):
        super().__init__()
        log.debug("Initializing VirtualLineList")
        self.parent: Org | Location = parent
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        log.debug("_get_data() started")
        params = {}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as data filter")
            params['orgId'] = self.parent.id
        elif isinstance(self.parent, wxcadm.Location):
            log.debug(f"Using Location ID {self.parent.id} as data filter")
            params['locationId'] = self.parent.id
        else:
            log.warn("Parent class is not Org or Location, so all items will be returned")
        response = webex_api_call('get', self._endpoint, params=params)
        items = []
        if self._endpoint_items_key is not None:
            log.info(f"Found {len(response[self._endpoint_items_key])} items")
            for entry in response[self._endpoint_items_key]:
                items.append(self._item_class(parent=self.parent, config=entry))
        else:
            log.info(f"Found {len(response)} items")
            for entry in response:
                items.append(self._item_class(parent=self.parent, config=entry))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self,
            id: Optional[str] = None,
            first_name: Optional[str] = None,
            last_name: Optional[str] = None,
            name: Optional[str] = None) -> Optional[VirtualLine]:
        """ Get the instance associated with a given ID or name.

        When searching by name, both `first_name` and `last_name` are required. Webex does not have a single key value
        other than ID, but the combination of first and last name must be unique. For names with only a single first
        name and single last name, such as 'Joe Smith', the ``name`` argument can be used as a shortcut.

        Args:
            id (str, optional): The Virtual Line ID to find
            first_name (str, optional): The first name of the Virtual Line
            last_name (str, optional): The last name of the Virtual Line
            name (str, optional): The full name (first and last) of the Virtual Line

        Returns:
            VirtualLine: The VirtualLine instance correlating to the given search argument. None is returned if no
                match is found

        Raises:
            ValueError: Raised when the method is called with no arguments or when called with only one name field

        """
        if id is not None:
            for item in self.data:
                if item.id == id:
                    return item
        if first_name is not None or last_name is not None or name is not None:
            if name is not None and first_name is None and last_name is None:
                first_name, last_name = name.split(' ', 1)
            if first_name and last_name:
                for item in self.data:
                    if item.first_name.lower() == first_name.lower() and item.last_name.lower() == last_name.lower():
                        return item
            else:
                raise ValueError("Both first_name and last_name are required to get by name")
        return None

    def create(
            self,
            first_name: str,
            last_name: str,
            phone_number: Optional[str] = None,
            extension: Optional[str] = None,
            location: Union[Location, str, None] = None,
            caller_id_first_name: Optional[str] = None,
            caller_id_last_name: Optional[str] = None,
            caller_id_number: Optional[str] = None
    ):
        """ Add a new Virtual Line

        When the :attr:`virtual_lines` is obtained at the :class:`Org` level, a ``location`` argument must be passed.
        When the :attr:`virtual_lines` is obtained at the :class:`Location` level, the ``location`` argument is
        not needed and assumed to be the :class:`Location` of the :attr:`virtual_lines`.

        Args:
            first_name (str): The first name of the Virtual Line
            last_name (str): The last name of the Virtual Line
            phone_number (str, optional): The phone number (DID) of the Virtual Line
            extension (str, optional): The extension of the Virtual Line
            location (Location, str, optional): The Location ID (str) or :class:`Location` instance
            caller_id_first_name (str, optional): The first name to user for Caller ID, if different from `first_name`
            caller_id_last_name (str, optional): The last name to use for Caller ID, if different from `last_name`
            caller_id_number (str, optional): The phone number to use for Caller ID if different from `phone_number`

        Returns:
            VirtualLine: The :class:`VirtualLine` that was added

        Raises:
            ValueError: Raised when both `extension` and `phone_number` are None. One or both are required
            APIError: Raised when the Virtual Line creation is rejected by the Webex API

        """
        log.info(f"Adding a Virtual Line to {self.parent.name}")
        # Make sure they passed a phone number or extension
        if phone_number is None and extension is None:
            raise ValueError("A phone_number, extension or both is required.")

        # Determine the location to add the VL to
        # If the List is at the Location level, use that Location
        if isinstance(self.parent, wxcadm.Location):
            log.debug(f"Using {self.parent.name} as target Location")
            location = self.parent
        # Get the Location ID to use for the payload
        if isinstance(location, wxcadm.Location):
            log.debug(f"Using {location.name} as Target Location")
            location_id = location.id
        elif isinstance(location, str):
            log.debug(f"Using {location} as target Location")
            location_id = location
        else:
            log.warn(f"Cannot determine target location using {location}")
            raise ValueError("Location or str required for location argument")
        # Build the payload
        payload = {
            'firstName': first_name,
            'lastName': last_name,
            'phoneNumber': phone_number,
            'extension': extension,
            'locationId': location_id,
            'callerIdFirstName': caller_id_first_name,
            'callerIdLastName': caller_id_last_name,
            'callerIdNumber': caller_id_number
        }

        response = webex_api_call('post', self._endpoint, payload=payload)
        new_entry_id = response['id']
        new_entry = self._item_class(self.parent, config={'id': new_entry_id})
        return new_entry



from __future__ import annotations

from typing import Optional
from datetime import datetime, timedelta, timezone
import wxcadm
from wxcadm import log
from .common import *


class Calls:
    def __init__(self, parent: wxcadm.Org):
        self.parent = parent

    def cdr(self, start: Optional[str] = None, end: Optional[str] = None, days: Optional[int] = None,
            hours: Optional[int] = None):
        """ Get a list of Call Detail Records

        Args:
            start (str, optional): The first date to include (YYYY-MM-DD)
            end (str, optional): The last date to include (YYYY-MM-DD)
            days (int, optional): The number of days to include, including today. Currently only 1 is supported by the
                API.
            hours (int, optional): The number of hours to include

        Returns:
            str: The Report ID

        .. note::
            This method requires a token with the ``spark-admin:calling_cdr_read`` scope. Additionally, the user
            must have the "Webex Calling Detailed Call History API access" option enabled in Control Hub, under the
            Administrator Roles.

        """
        log.debug('Calls.cdr() started')
        if start is None and days is None and hours is None:
            log.warning("No arguments provided")
            raise ValueError("You must provide either a start date, the number of day or the number of hours")
        if start is not None and (days is not None or hours is not None):
            log.warning("Start date and days/hours was passed. Invalid")
            raise ValueError("You must not specify a start date and number of days or hours")
        # Get the current UTC time to use if we need it
        timenow = datetime.now(timezone.utc)
        # Use "5 minutes ago" as the end time if we weren't given an end

        if end is None:
            endtime = timenow - timedelta(minutes=5)
            end = datetime.strftime(endtime, '%Y-%m-%dT%H:%M:%S.000Z')
            log.debug(f"No end provided. Using {end} as end time.")

        if hours is not None:
            # Fix for time sync issue with API. The API is very picky about the start time not being even a second
            # longer than 48 hours
            if hours == 48:
                starttime = timenow - timedelta(hours=hours) + timedelta(minutes=1)
            else:
                starttime = timenow - timedelta(hours=hours)
            start = datetime.strftime(starttime, '%Y-%m-%dT%H:%M:%S.000Z')
        elif days is not None:
            # Fix for time sync issue with API. The API is very picky about the start time not being even a second
            # longer than 48 hours
            if days == 2:
                starttime = timenow - timedelta(days=days) + timedelta(minutes=1)
            else:
                starttime = timenow - timedelta(days=days)
            start = datetime.strftime(starttime, '%Y-%m-%dT%H:%M:%S.000Z')
        elif start is not None:
            start = start

        log.debug(f'Setting start time to {start}')
        payload = {'startTime': start, 'endTime': end}
        response = webex_api_call('get', '/v1/cdr_feed', params=payload, domain='https://analytics.webexapis.com')
        return response



from __future__ import annotations

from collections import UserList

import wxcadm.location
from wxcadm import log
from .common import *


class PickupGroupList(UserList):
    def __init__(self, parent: wxcadm.Location):
        super().__init__()
        self.parent: wxcadm.Location = parent
        self.data: list = self._get_items()

    def _get_items(self):
        if isinstance(self.parent, wxcadm.Location):
            log.debug("Using Location as data filter")
        else:
            raise ValueError("Unsupported parent class")

        log.debug("Getting Call Pickup list")
        response = webex_api_call('get', f'v1/telephony/config/locations/{self.parent.id}/callPickups')
        log.debug(f"Received {len(response['callPickups'])} entries")
        items = []
        for entry in response['callPickups']:
            items.append(PickupGroup(self.parent, id=entry['id'], config=entry))
        return items

    def get(self, id: str = None, name: str = None, spark_id: str = None):
        """ Get the PickupGroup instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Auto
        Attendants will be searched in order by ID, Name, and finally Spark ID. If no arguments are provided, the method
        will raise an Exception.

        Args:
            id (str, optional): The AutoAttendant ID to find
            name (str, optional): The AutoAttendant Name to find
            spark_id (str, optional): The Spark ID to find

        Returns:
            AutoAttendant: The AutoAttendant instance correlating to the given search argument.
                None is returned if no AutoAttendant is found.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and spark_id is None:
            raise ValueError("A search argument must be provided")
        for aa in self.data:
            if aa.id == id:
                return aa
        for aa in self.data:
            if aa.name == name:
                return aa
        for aa in self.data:
            if aa.spark_id == spark_id:
                return aa
        return None


class PickupGroup:
    def __init__(self, parent: wxcadm.Location, id: str, config: dict):
        self.parent: wxcadm.Location = parent
        self.id: str = id
        """The Webex ID of the Pickup Group"""
        self.name: str = config['name']
        """The name of the Pickup Group"""
        self.config = config

    @property
    def users(self):
        """ All Users assigned to this Pickup Group"""
        log.info(f"Getting users for PickupGroups {self.name}")
        users = []
        response = webex_api_call('get', f'v1/telephony/config/locations/{self.parent.id}/callPickups/{self.id}')
        for item in response['agents']:
            #TODO - Once the Workspaces API supports Calling WOrkspaces, this should probably tie each Person/Workspace
            # to its corresponding instance
            users.append(item)
        return users



import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice


class TestDeviceList(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_devicelist_org(self):
        device_list = self.webex.org.devices
        self.assertIsInstance(device_list, wxcadm.device.DeviceList)
        success = device_list.refresh()
        self.assertTrue(success)

    def test_devicelist_person(self):
        person = choice(self.webex.org.people.webex_calling())
        device_list = person.devices
        self.assertIsInstance(device_list, wxcadm.device.DeviceList)

    def test_devicelist_workspace(self):
        workspace = choice(self.webex.org.workspaces.webex_calling())
        device_list = workspace.devices
        self.assertIsInstance(device_list, wxcadm.device.DeviceList)

    def test_devicelist_status(self) -> None:
        with self.subTest("Online Devices"):
            online_devices = self.webex.org.devices.get_by_status('online')
            if len(online_devices) == 0:
                self.skipTest("No online devices found")
            for device in online_devices:
                self.assertIn(device.connection_status, ['connected', 'connected_with_issues'])
        with self.subTest("Offline Devices"):
            offline_devices = self.webex.org.devices.get_by_status('offline')
            if len(offline_devices) == 0:
                self.skipTest("No offline devices found")
            for device in offline_devices:
                self.assertIn(device.connection_status, ['disconnected', 'offline_deep_sleep', 'offline_expired'])
        with self.subTest("Unknown Devices"):
            unknown_devices = self.webex.org.devices.get_by_status('unknown')
            if len(unknown_devices) == 0:
                self.skipTest("No unknown devices found")
            for device in unknown_devices:
                self.assertIn(device.connection_status, ['unknown'])



class TestDevice(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_device_members(self):
        device = choice(self.webex.org.devices.webex_calling())
        self.assertIsInstance(device, wxcadm.device.Device)
        members = device.members
        self.assertIsInstance(members, wxcadm.device.DeviceMemberList)

    def test_device_workspace_location_id(self):
        device = choice(self.webex.org.devices.webex_calling())
        self.assertIsNotNone(device.workspace_location_id)
        # The following test ensures that the .workspace_location_id can be decoded and matched to the WorkspaceLocation

class TestSupportedDevices(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)

    def test_supported_device_list(self):
        device_list = self.webex.org.devices.supported_devices
        self.assertIsInstance(device_list, wxcadm.device.SupportedDeviceList)
        with self.subTest("Devices in SupportedDeviceList"):
            self.assertGreater(len(device_list), 0)
        with self.subTest("SupportedDevice typing"):
            device = choice(device_list)
            self.assertIsInstance(device, wxcadm.device.SupportedDevice)

    def test_supported_device_get(self) -> None:
        # These test cases ensure that a SupportedDevice can be retrieved by name
        cisco_8851 = self.webex.org.supported_devices.get('8851')
        self.assertIsInstance(cisco_8851, wxcadm.SupportedDevice)
        self.assertEqual(cisco_8851.managed_by, 'CISCO')
        generic_phone = self.webex.org.supported_devices.get('generic ipphone')
        self.assertIsInstance(generic_phone, wxcadm.SupportedDevice)
        self.assertEqual(generic_phone.managed_by, 'CUSTOMER')


if __name__ == '__main__':
    unittest.main()